# -*- coding: utf-8 -*-
#
# Kivy documentation build configuration file, created by
# sphinx-quickstart on Wed Jan 21 22:37:12 2009.
#
# This file is execfile()d with the current directory set to its containing
# dir.
#
# The contents of this file are pickled, so don't put values in the namespace
# that aren't pickleable (module imports are okay, they're removed
# automatically).
#
# All configuration values have a default value; values that are commented out
# serve to show the default value.

import os
import sys
import configparser
import sphinx

# If your extensions are in another directory, add it here. If the directory
# is relative to the documentation root, use os.path.abspath to make it
# absolute, like shown here.
sys.path.insert(0, os.path.abspath('sphinxext'))
base_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, os.path.dirname(base_dir))

# General configuration
# ---------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'autodoc', 'sphinx.ext.todo', 'preprocess', 'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode', 'sphinx.ext.mathjax', 'sphinx.ext.extlinks',
    'sphinxcontrib.jquery']

if sphinx.version_info[0] >= 4:
    # In 4.0 and above has been added the support to substitute by ‘%s’ in the caption.
    # In 6.0 if the caption is a string, it must contain %s exactly once.
    repo_extlink_caption = '#%s'
else:
    repo_extlink_caption = '#'

extlinks = {'repo': ('https://github.com/kivy/kivy/issues/%s', repo_extlink_caption)}

# Todo configuration
todo_include_todos = True

# XXX HACK mathieu: monkey patch the autodoc module, to give a better priority
# for ClassDocumenter, or the cython class will be documented as AttributeClass
import sphinx.ext.autodoc
sphinx.ext.autodoc.ClassDocumenter.priority = 10

# Add any paths that contain templates here, relative to this directory.
if os.environ.get('READTHEDOCS') == 'True':
    templates_path = ['_templates']
else:
    templates_path = ['.templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General substitutions.
project = 'Kivy'
copyright = '2010-2024, Kivy Team and other contributors'

# The default replacements for |version| and |release|, also used in various
# other places throughout the built documents.
#
os.environ['KIVY_DOC_INCLUDE'] = '1'
import kivy
print(kivy.__file__)

version = kivy.__version__
release = kivy.__version__
base = 'autobuild.py-done'
if not os.path.exists(os.path.join(os.path.dirname(base_dir), base)):
    import autobuild
from kivy.tools import gallery
gallery.write_all_rst_pages()

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
today_fmt = '%B %d, %Y'

# suppress exclusion warnings
exclude_patterns = ['gsoc201*']

# The reST default role (used for this markup: `text`) to use for all documents
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'kivy_pygments_theme.KivyStyle'


# Options for HTML output
# -----------------------

# The style sheet to use for HTML and HTML Help pages. A file of that name
# must exist either in Sphinx' static/ path, or in one of the custom paths
# given in html_static_path.
html_style = 'fresh.css'

# Check for theme (remove 'if' when switched to RTD)
if os.environ.get('READTHEDOCS') == 'True':
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (within the static path) to place at the top of
# the sidebar.
html_logo = '.static/logo-kivy.png'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['.static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_use_modindex = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, the reST sources are included in the HTML build as _sources/<name>.
# html_copy_source = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# If nonempty, this is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = ''

# Output file base name for HTML help builder.
htmlhelp_basename = 'Kivydoc'


# Options for LaTeX output
# ------------------------

# The paper size ('letter' or 'a4').
# latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
# latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, document class [manual]).
latex_documents = [
  ('index', 'Kivy.tex', 'Kivy Documentation',
   'Kivy Team and other contributors', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

latex_elements = {
    'fontpkg':      r'\usepackage{mathpazo}',
    'papersize':    'a4paper',
    'pointsize':    '10pt',
    'preamble':     r'\usepackage{kivystyle}'
}
latex_additional_files = ['kivystyle.sty',
                          '../../kivy/data/logo/kivy-icon-512.png']

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# DEPRECATED! later_use_parts = True
# see sphinx.builder.latex.validate_config_values
latex_toplevel_sectioning = 'part'

# Additional stuff for the LaTeX preamble.
# latex_preamble = ''

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_use_modindex = True

config_parser = configparser.ConfigParser()
config_parser.read(os.path.join(base_dir, '..', '..', 'setup.cfg'))

# if used in a code-block, the block has to be marked with
# .. parse-literal::, otherwise it won't be replaced
# !!! doesn't work for "::", ".. code::" or ".. code-block::"
python_versions = config_parser['kivy']['python_versions'].strip()
cython_max_version = config_parser['kivy']['cython_max'].strip()
replacements = {
    'python_versions': python_versions,
    'kivy_version': kivy.__version__,
    'cython_install': f"Cython=={cython_max_version}",
    'python_versions_bold': f'**{python_versions}**',
    'kivy_version_bold': f'**{kivy.__version__}**',
}

epilog = []

for key, value in replacements.items():
    rep = '.. |{0}| replace:: {1}'.format(key, value)
    epilog.append(rep)

rst_epilog = '\n'.join(epilog)




__all__ = ('VertexFormat', 'VertexFormatException')

include "../include/config.pxi"
include "common.pxi"

from kivy.graphics.cgl cimport GL_FLOAT, GLfloat


class VertexFormatException(Exception):
    pass


cdef class VertexFormat:
    '''VertexFormat is used to describe the layout of the vertex data stored
    in vertex arrays/vbo's.

    .. versionadded:: 1.6.0
    '''
    def __cinit__(self, *fmt):
        self.vattr = NULL
        self.vattr_count = 0
        self.vsize = 0
        self.vbytesize = 0

    def __dealloc__(self):
        if self.vattr != NULL:
            free(self.vattr)
            self.vattr = NULL

    def __init__(self, *fmt):
        cdef vertex_attr_t *attr
        cdef int index, size

        if not fmt:
            raise VertexFormatException('No format specified')

        self.last_shader = None
        self.vattr_count = <long>len(fmt)
        self.vattr = <vertex_attr_t *>malloc(sizeof(vertex_attr_t) * self.vattr_count)

        if self.vattr == NULL:
            raise MemoryError()

        index = 0
        for name, size, tp in fmt:
            attr = &self.vattr[index]

            # fill the vertex format
            attr.per_vertex = 1
            attr.name = <bytes>name
            attr.index = 0 # will be set by the shader itself
            attr.size = size

            # only float is accepted as attribute format
            if tp == 'float':
                attr.type = GL_FLOAT
                attr.bytesize = sizeof(GLfloat) * size
            else:
                raise VertexFormatException('Unknown format type %r' % tp)

            # adjust the size, and prepare for the next iteration.
            index += 1
            self.vsize += attr.size
            self.vbytesize += attr.bytesize



"""
CGL/Glew: GL backend implementation using Glew
"""

include "../common.pxi"
include "../../include/config.pxi"

from kivy.graphics.cgl cimport *
from kivy.graphics.cgl_backend.cgl_gl import link_static
from kivy.logger import Logger

cdef extern from "gl_redirect.h":
    int glewInit() nogil
    int GLEW_OK
    char *glewGetErrorString(int) nogil

IF PLATFORM == "win32":
    cdef extern from *:
        void* wglGetProcAddress(const char* proc)

cpdef is_backend_supported():
    return not USE_OPENGL_MOCK and PLATFORM == "win32"


def init_backend():
    IF USE_OPENGL_MOCK or PLATFORM != "win32":
        raise TypeError('Glew is not available. Recompile with USE_OPENGL_MOCK=0')
    ELSE:
        cdef int result
        cdef bytes error
        result = glewInit()
        if result != GLEW_OK:
            error = glewGetErrorString(result)
            Logger.error('GL: GLEW initialization error {}'.format(error))
        else:
            Logger.info('GL: GLEW initialization succeeded')
        link_static()
        gl_dynamic_binding(<void *(__stdcall *)(const char *)>wglGetProcAddress)

IF not USE_OPENGL_MOCK and PLATFORM == "win32":
    cdef void gl_dynamic_binding(void *(__stdcall * f)(const char *)) except *:
        cdef bytes gl_extensions
        if cgl.glGetString == NULL:
            Logger.error('glGetString is unavailable, skipping Fbo detection')
            return
        gl_extensions = cgl.glGetString(GL_EXTENSIONS)

        # If the current opengl driver don't have framebuffers methods,
        # Check if an extension exist
        Logger.debug("GL: available extensions: {}".format(gl_extensions))
        if cgl.glGenFramebuffers != NULL:
            return

        Logger.debug("GL: glGenFramebuffers is NULL, try to detect an extension")
        if b"ARB_framebuffer_object" in gl_extensions:
            Logger.debug("GL: ARB_framebuffer_object is supported")

            cgl.glIsRenderbuffer = <GLISRENDERBUFFERPTR> f("glIsRenderbuffer")
            cgl.glBindRenderbuffer = <GLBINDRENDERBUFFERPTR> f("glBindRenderbuffer")
            cgl.glDeleteRenderbuffers = <GLDELETERENDERBUFFERSPTR> f("glDeleteRenderbuffers")
            cgl.glGenRenderbuffers = <GLGENRENDERBUFFERSPTR> f("glGenRenderbuffers")
            cgl.glRenderbufferStorage = <GLRENDERBUFFERSTORAGEPTR> f("glRenderbufferStorage")
            cgl.glGetRenderbufferParameteriv = <GLGETRENDERBUFFERPARAMETERIVPTR> f("glGetRenderbufferParameteriv")
            cgl.glIsFramebuffer = <GLISFRAMEBUFFERPTR> f("glIsFramebuffer")
            cgl.glBindFramebuffer = <GLBINDFRAMEBUFFERPTR> f("glBindFramebuffer")
            cgl.glDeleteFramebuffers = <GLDELETEFRAMEBUFFERSPTR> f("glDeleteFramebuffers")
            cgl.glGenFramebuffers = <GLGENFRAMEBUFFERSPTR> f("glGenFramebuffers")
            cgl.glCheckFramebufferStatus = <GLCHECKFRAMEBUFFERSTATUSPTR> f("glCheckFramebufferStatus")
            #cgl.glFramebufferTexture1D = <GLFRAMEBUFFERTEXTURE1DPTR> f("glFramebufferTexture1D")
            cgl.glFramebufferTexture2D = <GLFRAMEBUFFERTEXTURE2DPTR> f("glFramebufferTexture2D")
            #cgl.glFramebufferTexture3D = <GLFRAMEBUFFERTEXTURE3DPTR> f("glFramebufferTexture3D")
            cgl.glFramebufferRenderbuffer = <GLFRAMEBUFFERRENDERBUFFERPTR> f("glFramebufferRenderbuffer")
            cgl.glGetFramebufferAttachmentParameteriv = <GLGETFRAMEBUFFERATTACHMENTPARAMETERIVPTR> f("glGetFramebufferAttachmentParameteriv")
            cgl.glGenerateMipmap = <GLGENERATEMIPMAPPTR> f("glGenerateMipmap")
        elif b"EXT_framebuffer_object" in gl_extensions:
            Logger.debug("GL: EXT_framebuffer_object is supported\n")

            cgl.glIsRenderbuffer = <GLISRENDERBUFFERPTR> f("glIsRenderbufferEXT")
            cgl.glBindRenderbuffer = <GLBINDRENDERBUFFERPTR> f("glBindRenderbufferEXT")
            cgl.glDeleteRenderbuffers = <GLDELETERENDERBUFFERSPTR> f("glDeleteRenderbuffersEXT")
            cgl.glGenRenderbuffers = <GLGENRENDERBUFFERSPTR> f("glGenRenderbuffersEXT")
            cgl.glRenderbufferStorage = <GLRENDERBUFFERSTORAGEPTR> f("glRenderbufferStorageEXT")
            cgl.glGetRenderbufferParameteriv = <GLGETRENDERBUFFERPARAMETERIVPTR> f("glGetRenderbufferParameterivEXT")
            cgl.glIsFramebuffer = <GLISFRAMEBUFFERPTR> f("glIsFramebufferEXT")
            cgl.glBindFramebuffer = <GLBINDFRAMEBUFFERPTR> f("glBindFramebufferEXT")
            cgl.glDeleteFramebuffers = <GLDELETEFRAMEBUFFERSPTR> f("glDeleteFramebuffersEXT")
            cgl.glGenFramebuffers = <GLGENFRAMEBUFFERSPTR> f("glGenFramebuffersEXT")
            cgl.glCheckFramebufferStatus = <GLCHECKFRAMEBUFFERSTATUSPTR> f("glCheckFramebufferStatusEXT")
            #cgl.glFramebufferTexture1D = <GLFRAMEBUFFERTEXTURE1DPTR> f("glFramebufferTexture1DEXT")
            cgl.glFramebufferTexture2D = <GLFRAMEBUFFERTEXTURE2DPTR> f("glFramebufferTexture2DEXT")
            #cgl.glFramebufferTexture3D = <GLFRAMEBUFFERTEXTURE3DPTR> f("glFramebufferTexture3DEXT")
            cgl.glFramebufferRenderbuffer = <GLFRAMEBUFFERRENDERBUFFERPTR> f("glFramebufferRenderbufferEXT")
            cgl.glGetFramebufferAttachmentParameteriv = <GLGETFRAMEBUFFERATTACHMENTPARAMETERIVPTR> f("glGetFramebufferAttachmentParameterivEXT")
        else:
            Logger.info("GL: No framebuffers extension is supported")
            Logger.debug("GL: Any call to Fbo will crash!")



'''
Scissor Instructions
====================

.. versionadded:: 1.9.1


Scissor instructions clip your drawing area into a rectangular region.

- :class:`ScissorPush`: Begins clipping, sets the bounds of the clip space
- :class:`ScissorPop`: Ends clipping

The area provided to clip is in screenspace pixels and must be provided as
integer values not floats.

The following code will draw a circle on top of our widget while clipping
the circle so it does not expand beyond the widget borders.

.. code-block:: python

    with self.canvas.after:
        #If our widget is inside another widget that modified the coordinates
        #spacing (such as ScrollView) we will want to convert to Window coords
        x,y = self.to_window(*self.pos)
        width, height = self.size
        #We must convert from the possible float values provided by kivy
        #widgets to an integer screenspace, in python3 round returns an int so
        #the int cast will be unnecessary.
        ScissorPush(x=int(round(x)), y=int(round(y)),
            width=int(round(width)), height=int(round(height)))
        Color(rgba=(1., 0., 0., .5))
        Ellipse(size=(width*2., height*2.),
            pos=self.center)
        ScissorPop()
'''
include "../include/config.pxi"
include "opcodes.pxi"

from kivy.graphics.cgl cimport *
from kivy.graphics.instructions cimport Instruction

cdef class Rect:
    '''Rect class used internally by ScissorStack and ScissorPush to determine
    correct clipping area.
    '''
    cdef int _x
    cdef int _y
    cdef int _width
    cdef int _height

    def __init__(self, int x, int y, int width, int height):
        self._x = x
        self._y = y
        self._width = width
        self._height = height

    def intersect(self, Rect other):
        max_x = min(self._x + self._width, other._x + other._width)
        x = max(self._x, other._x)
        width = max(0, max_x - x)
        max_y = min(self._y + self._height, other._y + other._height)
        y = max(self._y, other._y)
        height = max(0, max_y - y)
        self._x = x
        self._width = width
        self._y = y
        self._height = height


cdef class ScissorStack:
    '''Class used internally to keep track of the current state of
    glScissors regions. Do not instantiate, prefer to inspect the module's
    scissor_stack.
    '''
    cdef list _stack

    def __init__(self):
        self._stack = []

    @property
    def empty(self):
        return True if len(self._stack) is 0 else False

    @property
    def back(self):
        return self._stack[-1]

    def push(self, element):
        self._stack.append(element)

    def pop(self):
        return self._stack.pop()


scissor_stack = ScissorStack()


cdef class ScissorPush(Instruction):
    '''Push the scissor stack. Provide kwargs of 'x', 'y', 'width', 'height'
    to control the area and position of the scissoring region. Defaults to
    0, 0, 100, 100

    Scissor works by clipping all drawing outside of a rectangle starting at
    int x, int y position and having sides of int width by int height in Window
    space coordinates
    '''
    cdef int _x
    cdef int _y
    cdef int _width
    cdef int _height
    cdef Rect _rect

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, value):
        self._x = value
        self._rect = Rect(self._x, self._y, self._width, self._height)
        self.flag_data_update()

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, value):
        self._y = value
        self._rect = Rect(self._x, self._y, self._width, self._height)
        self.flag_data_update()

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, value):
        self._width = value
        self._rect = Rect(self._x, self._y, self._width, self._height)
        self.flag_data_update()

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, value):
        self._height = value
        self._rect = Rect(self._x, self._y, self._width, self._height)
        self.flag_data_update()

    @property
    def pos(self):
        return self._x, self._y

    @pos.setter
    def pos(self, value):
        self._x, self._y = value
        self._rect = Rect(self._x, self._y, self._width, self._height)
        self.flag_data_update()

    @property
    def size(self):
        return self._width, self._height

    @size.setter
    def size(self, value):
        self._width, self._height = value
        self._rect = Rect(self._x, self._y, self._width, self._height)
        self.flag_data_update()

    def __init__(self, **kwargs):
        self._x, self._y = kwargs.pop(
            'pos', (
                kwargs.pop('x', 0),
                kwargs.pop('y', 0)
                )
            )
        self._width, self._height = kwargs.pop(
            'size', (
                kwargs.pop('width', 100),
                kwargs.pop('height', 100)
                )
            )
        super(ScissorPush, self).__init__(**kwargs)
        self._rect = Rect(self._x, self._y, self._width, self._height)

    cdef int apply(self) except -1:
        cdef Rect rect = self._rect
        cdef Rect new_scissor_rect
        cdef Rect back
        if scissor_stack.empty:
            scissor_stack.push(rect)
            cgl.glEnable(GL_SCISSOR_TEST)
            cgl.glScissor(self._x, self._y, self._width, self._height)
        else:
            new_scissor_rect = Rect(rect._x, rect._y,
                rect._width, rect._height)
            back = scissor_stack.back
            new_scissor_rect.intersect(back)
            scissor_stack.push(new_scissor_rect)
            cgl.glScissor(new_scissor_rect._x, new_scissor_rect._y,
                new_scissor_rect._width, new_scissor_rect._height)

cdef class ScissorPop(Instruction):
    '''Pop the scissor stack. Call after ScissorPush, once you have completed
    the drawing you wish to be clipped.
    '''

    cdef int apply(self) except -1:
        scissor_stack.pop()
        cdef Rect new_scissor_rect
        if scissor_stack.empty:
            cgl.glDisable(GL_SCISSOR_TEST)
        else:
            new_scissor_rect = scissor_stack.back
            cgl.glScissor(new_scissor_rect._x, new_scissor_rect._y,
                new_scissor_rect._width, new_scissor_rect._height)



'''
ImageIO OSX framework
=====================

Image loader implementation based on CoreGraphics OSX framework.

.. todo::

    clean all unused definitions
    handle all errors cases
    add documentation

'''

__all__ = ('ImageLoaderImageIO', )

from kivy.logger import Logger
from kivy.core.image import ImageLoaderBase, ImageData, ImageLoader

from array import array
from libcpp cimport bool
from libc.stdlib cimport malloc, free
from libc.string cimport memcpy

ctypedef unsigned long size_t
ctypedef signed long CFIndex

cdef unsigned int kCFStringEncodingUTF8 = 0x08000100

cdef extern from "stdlib.h" nogil:
    void* calloc(size_t, size_t)

cdef extern from "CoreGraphics/CGDataProvider.h" nogil:
    ctypedef void *CFDataRef
    unsigned char *CFDataGetBytePtr(CFDataRef)

    ctypedef struct CGPoint:
        float x
        float y

    ctypedef struct CGSize:
        float width
        float height

    ctypedef struct CGRect:
        CGPoint origin
        CGSize size

    CGRect CGRectMake(float, float, float, float)

cdef extern from "CoreFoundation/CFBase.h" nogil:
    ctypedef void *CFAllocatorRef
    ctypedef void *CFStringRef
    ctypedef void *CFTypeRef
    CFStringRef CFStringCreateWithCString(CFAllocatorRef alloc, char *cStr,
            int encoding)

    void CFRelease(CFTypeRef cf)

cdef extern from "CoreFoundation/CFURL.h" nogil:
    ctypedef void *CFURLRef
    ctypedef int CFURLPathStyle
    int kCFURLPOSIXPathStyle
    CFURLRef CFURLCreateFromFileSystemRepresentation(
            CFAllocatorRef, unsigned char *, CFIndex, bool)
    CFURLRef CFURLCreateWithFileSystemPath(CFAllocatorRef allocator,
            CFStringRef filePath, CFURLPathStyle pathStyle, int isDirectory)

cdef extern from "CoreFoundation/CFDictionary.h":
    ctypedef void *CFDictionaryRef

cdef extern from "CoreFoundation/CoreFoundation.h" nogil:
    CFDataRef CFDataCreateWithBytesNoCopy(
                CFAllocatorRef, const unsigned char *, int length,
                CFAllocatorRef)

cdef extern from "CoreGraphics/CGImage.h" nogil:
    ctypedef void *CGImageRef
    void CGImageRelease(CGImageRef image)
    size_t CGImageGetWidth(CGImageRef)
    size_t CGImageGetHeight(CGImageRef)
    int kCGImageAlphaNoneSkipLast
    int kCGImageAlphaNoneSkipFirst
    int kCGImageAlphaFirst
    int kCGImageAlphaLast
    int kCGImageAlphaPremultipliedLast
    int kCGImageAlphaPremultipliedFirst
    int kCGBitmapByteOrder32Host

cdef extern from "CoreGraphics/CGColorSpace.h" nogil:
    ctypedef void *CGColorSpaceRef
    CGColorSpaceRef CGColorSpaceCreateDeviceRGB()
    void CGColorSpaceRelease(CGColorSpaceRef cs)

cdef extern from "CoreGraphics/CGAffineTransform.h" nogil:
    ctypedef void *CGAffineTransform
    CGAffineTransform CGAffineTransformMake(float a, float b, float c, float d, float tx, float ty)

cdef extern from "CoreGraphics/CGContext.h" nogil:
    ctypedef void *CGContextRef
    void CGContextRelease(CGContextRef c)
    void CGContextDrawImage(CGContextRef, CGRect, CGImageRef)
    int kCGBlendModeCopy
    int kCGBlendModeNormal
    void CGContextSetBlendMode(CGContextRef, int)
    void CGContextConcatCTM(CGContextRef fc, CGAffineTransform matrix)

cdef extern from "CoreGraphics/CGBitmapContext.h" nogil:
    CGImageRef CGBitmapContextCreateImage(CGColorSpaceRef)
    CGContextRef CGBitmapContextCreate(
       void *data, size_t width, size_t height, size_t bitsPerComponent,
       size_t bytesPerRow, CGColorSpaceRef colorspace, unsigned int bitmapInfo)

cdef extern from "ImageIO/CGImageSource.h" nogil:
    ctypedef void *CGImageSourceRef
    CGImageSourceRef CGImageSourceCreateWithURL(
            CFURLRef, CFDictionaryRef)
    CGImageRef CGImageSourceCreateImageAtIndex(
            CGImageSourceRef, size_t, CFDictionaryRef)
    CGImageRef CGImageSourceCreateWithData(
            CFDataRef data, CFDictionaryRef options)

cdef extern from "ImageIO/CGImageDestination.h" nogil:
    ctypedef void *CGImageDestinationRef
    CGImageDestinationRef CGImageDestinationCreateWithURL(
        CFURLRef, CFStringRef, size_t, CFDictionaryRef)
    void CGImageDestinationAddImage(CGImageDestinationRef idst,
        CGImageRef image, CFDictionaryRef properties)
    int CGImageDestinationFinalize(CGImageDestinationRef idst)

cdef extern from "Accelerate/Accelerate.h" nogil:
    ctypedef struct vImage_Buffer:
        void *data
        int width
        int height
        size_t rowBytes

    int vImagePermuteChannels_ARGB8888(
            vImage_Buffer *src, vImage_Buffer *dst, unsigned char *permuteMap,
            int flags)


def load_image_data(bytes _url, bytes _data=None):
    cdef size_t width, height
    cdef char *r_data = NULL
    cdef size_t datalen = 0
    cdef char *c_url = NULL
    cdef char *c_data = NULL
    if _url:
        c_url = _url
        datalen = len(_url)
    if _data:
        c_data = _data
        datalen = len(_data)
    c_load_image_data(c_url, c_data, datalen, &width, &height, &r_data)
    if r_data == NULL:
        raise ValueError("No image to load at {}".format(_url))
    py_data = r_data[:width * height * 4]
    free(r_data)
    return (width, height, 'rgba', py_data)


cdef void c_load_image_data(char *_url, char *_data, size_t datalen, size_t *width, size_t
        *height, char **r_data) nogil:
    # load an image from the _url with CoreGraphics, and output an RGBA string.
    cdef CFURLRef url = NULL
    cdef CGImageSourceRef myImageSourceRef
    cdef CFDataRef dataref = NULL

    width[0] = height[0] = 0
    r_data[0] = NULL

    if _data != NULL:
        dataref = CFDataCreateWithBytesNoCopy(
            NULL, <const unsigned char*>_data, datalen, NULL)
        myImageSourceRef = CGImageSourceCreateWithData(dataref, NULL)
        if not myImageSourceRef:
            CFRelease(dataref)
            return
    else:
        url = CFURLCreateFromFileSystemRepresentation(NULL, <unsigned char *>_url, datalen, 0)
        myImageSourceRef = CGImageSourceCreateWithURL(url, NULL)
        if not myImageSourceRef:
            CFRelease(url)
            return
    cdef CGImageRef myImageRef = CGImageSourceCreateImageAtIndex(myImageSourceRef, 0, NULL)
    width[0] = CGImageGetWidth(myImageRef)
    height[0] = CGImageGetHeight(myImageRef)
    if myImageRef == NULL:
        CFRelease(myImageSourceRef)
        return
    cdef CGRect rect = CGRectMake(0, 0, width[0], height[0])
    cdef CGColorSpaceRef space = CGColorSpaceCreateDeviceRGB()
    cdef vImage_Buffer src
    cdef vImage_Buffer dest
    dest.height = src.height = height[0]
    dest.width = src.width = width[0]
    dest.rowBytes = src.rowBytes = width[0] * 4
    src.data = calloc(width[0] * 4, height[0])
    dest.data = r_data[0] = <char *>calloc(width[0] * 4, height[0])

    # endianness:  kCGBitmapByteOrder32Little = (2 << 12)
    # (2 << 12) | kCGImageAlphaPremultipliedLast)
    cdef CGContextRef myBitmapContext = CGBitmapContextCreate(
            src.data, width[0], height[0], 8, width[0] * 4, space,
            kCGBitmapByteOrder32Host | kCGImageAlphaNoneSkipFirst)

    CGContextSetBlendMode(myBitmapContext, kCGBlendModeCopy)
    CGContextDrawImage(myBitmapContext, rect, myImageRef)

    # convert to RGBA using Accelerate framework
    cdef unsigned char *pmap = [2, 1, 0, 3]
    vImagePermuteChannels_ARGB8888(&src, &dest, pmap, 0)

    # release everything
    CGImageRelease(<CGImageRef>myImageSourceRef)
    CFRelease(myImageRef)
    CGContextRelease(myBitmapContext)
    CGColorSpaceRelease(space)
    free(src.data)
    #free(dest.data)  # this part is freed by the caller.

def save_image_rgba(filename, width, height, data, flipped):
    # compatibility, could be removed i guess
    save_image(filename, width, height, 'rgba', data, flipped)


def save_image(filenm, width, height, fmt, data, flipped):
    # save a RGBA string into filename using CoreGraphics

    # FIXME only png output are accepted.
    # the day we want to support another output format, we need to adapt the
    # ctype variable: "public.png" is not a name, but a domain that represent
    # the type of the output file. So we need to map the extension of the
    # filename into a CoreGraphics image domain type.

    cdef bytes fileformat = b'public.png'
    cdef bytes filename = <bytes>filenm.encode('utf-8')
    if filename.endswith(b'.png'):
        fileformat = b'public.png'
    if filename.endswith(b'.jpg') or filename.endswith(b'.jpeg'):
        fileformat = b'public.jpeg'

    cdef char *source = NULL
    if type(data) is array:
        data = data.tostring()
    bsource = <bytes>data[:len(data)]
    source = <char *> bsource

    cdef int fmt_length = 3
    if fmt == 'rgba':
        fmt_length = 4
    cdef char *pixels = <char *>malloc(int(width * height * fmt_length))
    memcpy(pixels, <void *>source, int(width * height * fmt_length))

    cdef CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB()
    cdef CGContextRef bitmapContext = CGBitmapContextCreate(
        pixels, width, height,
        8, # bitsPerComponent
        fmt_length * width, # bytesPerRow
        colorSpace,
        kCGImageAlphaNoneSkipLast)

    cdef CGImageRef cgImage = CGBitmapContextCreateImage(bitmapContext)
    cdef char *cfilename = <char *>filename

    cdef CFStringRef sfilename = CFStringCreateWithCString(NULL,
            cfilename, kCFStringEncodingUTF8)
    cdef CFURLRef url = CFURLCreateWithFileSystemPath(NULL,
            sfilename, kCFURLPOSIXPathStyle, 0)
    cdef CFStringRef ctype = CFStringCreateWithCString(NULL,
            fileformat, kCFStringEncodingUTF8)

    cdef CGImageDestinationRef dest = CGImageDestinationCreateWithURL(url,
            ctype, 1, NULL)

    # copy the image into a transformed context
    cdef CGContextRef flippedContext
    cdef CGImageRef newImageRef

    if flipped:
        flippedContext = CGBitmapContextCreate(
                                    NULL, width, height,
                                    8, # bitsPerComponent
                                    fmt_length * width, # bytesPerRow
                                    colorSpace,
                                    kCGImageAlphaNoneSkipLast)

        CGContextConcatCTM(flippedContext, CGAffineTransformMake(1.0, 0.0,
                                                                0.0, -1.0,
                                                                0.0, height))

        CGContextDrawImage(flippedContext,
                            CGRectMake(0, 0, width, height),
                            cgImage)

        newImageRef = CGBitmapContextCreateImage(flippedContext)
        CGImageDestinationAddImage(dest, newImageRef, NULL)
        CGImageDestinationFinalize(dest)
        CFRelease(newImageRef)
        CFRelease(flippedContext)
    else:
        CGImageDestinationAddImage(dest, cgImage, NULL)
        CGImageDestinationFinalize(dest)

    # Release everything
    CFRelease(cgImage)
    CFRelease(bitmapContext)
    CFRelease(colorSpace)
    free(pixels)


class ImageLoaderImageIO(ImageLoaderBase):
    '''Image loader based on ImageIO OS X Framework
    '''

    @staticmethod
    def extensions():
        # FIXME check which one are available on osx
        return ('bmp', 'bufr', 'cur', 'dcx', 'fits', 'fl', 'fpx', 'gbr',
                'gd', 'grib', 'hdf5', 'ico', 'im', 'imt', 'iptc',
                'jpeg', 'jpg', 'jpe', 'mcidas', 'mic', 'mpeg', 'msp',
                'pcd', 'pcx', 'pixar', 'png', 'ppm', 'psd', 'sgi',
                'spider', 'tga', 'tiff', 'wal', 'wmf', 'xbm', 'xpm',
                'xv', 'icns')

    def load(self, filename):
        # FIXME: if the filename is unicode, the loader is failing.
        if self._inline:
            data = filename.read()
            ret = load_image_data(None, data)
        else:
            ret = load_image_data(filename.encode('utf-8'))
        if ret is None:
            Logger.warning('Image: Unable to load image <%s>' % filename)
            raise Exception('Unable to load image')
        w, h, imgtype, data = ret
        return [ImageData(w, h, imgtype, data, source=filename)]

    @staticmethod
    def can_save(fmt, is_bytesio):
        if is_bytesio:
            return False
        return fmt in ImageLoaderImageIO.extensions()

    @staticmethod
    def can_load_memory():
        return True

    @staticmethod
    def save(filename, width, height, pixelfmt, pixels, flipped=False,
             imagefmt=None):
        save_image(filename, width, height, pixelfmt, pixels, flipped)
        return True

# register
ImageLoader.register(ImageLoaderImageIO)



#
# Kivy - Cross-platform UI framework
# https://kivy.org/
#

import sys
build_examples = False
if "--build_examples" in sys.argv:
    build_examples = True
    sys.argv.remove("--build_examples")

from kivy.utils import pi_version
from copy import deepcopy
import os
from os.path import join, dirname, exists, basename, isdir
from os import walk, environ, makedirs
from collections import OrderedDict
from time import sleep
from pathlib import Path
import logging
import sysconfig
import textwrap
import tempfile

from setuptools import Distribution, Extension, find_packages, setup
from setuptools.command.build_ext import build_ext

if sys.version_info[0] == 2:
    logging.critical(
        'Unsupported Python version detected!: Kivy 2.0.0 and higher does not '
        'support Python 2. Please upgrade to Python 3, or downgrade Kivy to '
        '1.11.1 - the last Kivy release that still supports Python 2.')


def ver_equal(self, other):
    return self.version == other


def get_description():
    with open(join(dirname(__file__), 'README.md'), 'rb') as fileh:
        return fileh.read().decode("utf8").replace('\r\n', '\n')


def getoutput(cmd, env=None):
    import subprocess
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE, env=env)
    p.wait()
    if p.returncode:  # if not returncode == 0
        print('WARNING: A problem occurred while running {0} (code {1})\n'
              .format(cmd, p.returncode))
        stderr_content = p.stderr.read()
        if stderr_content:
            print('{0}\n'.format(stderr_content))
        return ""
    return p.stdout.read()


def pkgconfig(*packages, **kw):
    flag_map = {'-I': 'include_dirs', '-L': 'library_dirs', '-l': 'libraries'}
    lenviron = None

    if platform == 'win32':
        pconfig = join(sys.prefix, 'libs', 'pkgconfig')
        if isdir(pconfig):
            lenviron = environ.copy()
            lenviron['PKG_CONFIG_PATH'] = '{};{}'.format(
                environ.get('PKG_CONFIG_PATH', ''), pconfig)

    if KIVY_DEPS_ROOT and platform != 'win32':
        lenviron = environ.copy()
        lenviron["PKG_CONFIG_PATH"] = "{}:{}:{}".format(
            environ.get("PKG_CONFIG_PATH", ""),
            join(
                KIVY_DEPS_ROOT, "dist", "lib", "pkgconfig"
            ),
            join(
                KIVY_DEPS_ROOT, "dist", "lib64", "pkgconfig"
            ),
        )

    cmd = 'pkg-config --libs --cflags {}'.format(' '.join(packages))
    results = getoutput(cmd, lenviron).split()
    for token in results:
        ext = token[:2].decode('utf-8')
        flag = flag_map.get(ext)
        if not flag:
            continue
        kw.setdefault(flag, []).append(token[2:].decode('utf-8'))
    return kw


def get_isolated_env_paths():
    try:
        # sdl2_dev is installed before setup.py is run, when installing from
        # source due to pyproject.toml. However, it is installed to a
        # pip isolated env, which we need to add to compiler
        import kivy_deps.sdl2_dev as sdl2_dev
    except ImportError:
        return [], []

    root = os.path.abspath(join(sdl2_dev.__path__[0], '../../../..'))
    includes = [join(root, 'Include')] if isdir(join(root, 'Include')) else []
    libs = [join(root, 'libs')] if isdir(join(root, 'libs')) else []
    return includes, libs


def check_c_source_compiles(code, include_dirs=None):
    """Check if C code compiles.
    This function can be used to check if a specific feature is available on
    the current platform, and therefore enable or disable some modules.
    """

    def get_compiler():
        """Get the compiler instance used by setuptools.
        This is a bit hacky, but seems the only way to get the compiler instance
        used by setuptools, without using private APIs or the deprecated
        distutils module. (See: https://github.com/pypa/setuptools/issues/2806)
        """
        fake_dist_build_ext = Distribution().get_command_obj("build_ext")
        fake_dist_build_ext.finalize_options()
        # register an extension to ensure a compiler is created
        fake_dist_build_ext.extensions = [Extension("ignored", ["ignored.c"])]
        # disable building fake extensions
        fake_dist_build_ext.build_extensions = lambda: None
        # run to populate self.compiler
        fake_dist_build_ext.run()
        return fake_dist_build_ext.compiler

    # Create a temporary file which contains the code
    with tempfile.TemporaryDirectory() as tmpdir:
        temp_file = os.path.join(tmpdir, "test.c")
        with open(temp_file, "w", encoding="utf-8") as tf:
            tf.write(code)
        try:
            get_compiler().compile(
                [temp_file], extra_postargs=[], include_dirs=include_dirs
            )
        except Exception as ex:
            print(ex)
            return False
    return True


# -----------------------------------------------------------------------------

# Determine on which platform we are

build_examples = build_examples or \
    os.environ.get('KIVY_BUILD_EXAMPLES', '0') == '1'

platform = sys.platform

# Detect Python for android project (http://github.com/kivy/python-for-android)
ndkplatform = environ.get('NDKPLATFORM')
if ndkplatform is not None and environ.get('LIBLINK'):
    platform = 'android'
kivy_ios_root = environ.get('KIVYIOSROOT', None)
if kivy_ios_root is not None:
    platform = 'ios'
# proprietary broadcom video core drivers
if exists('/opt/vc/include/bcm_host.h'):
    used_pi_version = pi_version
    # Force detected Raspberry Pi version for cross-builds, if needed
    if 'KIVY_RPI_VERSION' in environ:
        used_pi_version = int(environ['KIVY_RPI_VERSION'])
    # The proprietary broadcom video core drivers are not available on the
    # Raspberry Pi 4
    if (used_pi_version or 4) < 4:
        platform = 'rpi'
# use mesa video core drivers
if environ.get('VIDEOCOREMESA', None) == '1':
    platform = 'vc'
mali_paths = (
    '/usr/lib/arm-linux-gnueabihf/libMali.so',
    '/usr/lib/arm-linux-gnueabihf/mali-egl/libmali.so',
    '/usr/local/mali-egl/libmali.so')
if any((exists(path) for path in mali_paths)):
    platform = 'mali'

# Needed when cross-compiling
if environ.get('KIVY_CROSS_PLATFORM'):
    platform = environ.get('KIVY_CROSS_PLATFORM')

# If the user has specified a KIVY_DEPS_ROOT, use that as the root for
# (ATM only SDL) dependencies. Otherwise, use the default locations.
KIVY_DEPS_ROOT = os.environ.get('KIVY_DEPS_ROOT', None)

# if KIVY_DEPS_ROOT is None and platform is linux or darwin show a warning
# message, because using a system provided SDL2 is not recommended.
# (will be shown only in verbose mode)
if KIVY_DEPS_ROOT is None and platform in ('linux', 'darwin'):
    print("###############################################")
    print("WARNING: KIVY_DEPS_ROOT is not set, using system provided SDL")
    print("which is not recommended as it may be incompatible with Kivy.")
    print("Please build dependencies from source via the provided script")
    print("and set KIVY_DEPS_ROOT to the root of the dependencies directory.")
    print("###############################################")


# -----------------------------------------------------------------------------
# Detect options
#
c_options = OrderedDict()
c_options['use_rpi_vidcore_lite'] = platform == 'rpi'
c_options['use_egl'] = False
c_options['use_opengl_es2'] = None
c_options['use_opengl_mock'] = environ.get('READTHEDOCS', None) == 'True'
c_options['use_sdl2'] = None
c_options['use_pangoft2'] = None
c_options['use_ios'] = False
c_options['use_android'] = False
c_options['use_mesagl'] = False
c_options['use_x11'] = False
c_options['use_wayland'] = False
c_options['use_gstreamer'] = None
c_options['use_avfoundation'] = platform in ['darwin', 'ios']
c_options['use_osx_frameworks'] = platform == 'darwin'
c_options['debug_gl'] = False

# Set the alpha size, this will be 0 on the Raspberry Pi and 8 on all other
# platforms, so SDL2 works without X11
c_options['kivy_sdl_gl_alpha_size'] = 8 if pi_version is None else 0

# now check if environ is changing the default values
for key in list(c_options.keys()):
    ukey = key.upper()
    if ukey in environ:
        # kivy_sdl_gl_alpha_size should be an integer, the rest are booleans
        value = int(environ[ukey])
        if key != 'kivy_sdl_gl_alpha_size':
            value = bool(value)
        print('Environ change {0} -> {1}'.format(key, value))
        c_options[key] = value

use_embed_signature = environ.get('USE_EMBEDSIGNATURE', '0') == '1'
use_embed_signature = use_embed_signature or bool(
    platform not in ('ios', 'android'))

# -----------------------------------------------------------------------------
# We want to be able to install kivy as a wheel without a dependency
# on cython, but we also want to use cython where possible as a setup
# time dependency through `pyproject.toml` if building from source.

# There are issues with using cython at all on some platforms;
# exclude them from using or declaring cython.

# This determines whether Cython specific functionality may be used.
can_use_cython = True

if platform in ('ios', 'android'):
    # NEVER use or declare cython on these platforms
    print('Not using cython on %s' % platform)
    can_use_cython = False


# -----------------------------------------------------------------------------
# Setup classes

# the build path where kivy is being compiled
src_path = build_path = dirname(__file__)
print("Current directory is: {}".format(os.getcwd()))
print("Source and initial build directory is: {}".format(src_path))

# __version__ is imported by exec, but help linter not complain
__version__ = None
with open(join(src_path, 'kivy', '_version.py'), encoding="utf-8") as f:
    exec(f.read())


class KivyBuildExt(build_ext, object):

    def finalize_options(self):
        super().finalize_options()

        # Build the extensions in parallel if the options has not been set
        if hasattr(self, 'parallel') and self.parallel is None:
            # Use a maximum of 4 cores. If cpu_count returns None, then parallel
            # build will be disabled
            self.parallel = min(4, os.cpu_count() or 0)
            if self.parallel:
                print('Building extensions in parallel using {} cores'.format(
                    self.parallel))

        global build_path
        if (self.build_lib is not None and exists(self.build_lib) and
                not self.inplace):
            build_path = self.build_lib
            print("Updated build directory to: {}".format(build_path))

    def build_extensions(self):
        # build files
        config_h_fn = ('include', 'config.h')
        config_pxi_fn = ('include', 'config.pxi')
        config_py_fn = ('setupconfig.py', )

        # generate headers
        config_h = '// Autogenerated file for Kivy C configuration\n'
        config_h += '#define __PY3 1\n'
        config_pxi = '# Autogenerated file for Kivy Cython configuration\n'
        config_pxi += 'DEF PY3 = 1\n'
        config_py = '# Autogenerated file for Kivy configuration\n'
        config_py += 'PY3 = 1\n'
        config_py += 'CYTHON_MIN = {0}\nCYTHON_MAX = {1}\n'.format(
            repr(MIN_CYTHON_STRING), repr(MAX_CYTHON_STRING))
        config_py += 'CYTHON_BAD = {0}\n'.format(repr(', '.join(map(
            str, CYTHON_UNSUPPORTED))))

        # generate content
        print('Build configuration is:')
        for opt, value in c_options.items():
            # kivy_sdl_gl_alpha_size is already an integer
            if opt != 'kivy_sdl_gl_alpha_size':
                value = int(bool(value))
            print(' * {0} = {1}'.format(opt, value))
            opt = opt.upper()
            config_h += '#define __{0} {1}\n'.format(opt, value)
            config_pxi += 'DEF {0} = {1}\n'.format(opt, value)
            config_py += '{0} = {1}\n'.format(opt, value)
        debug = bool(self.debug)
        print(' * debug = {0}'.format(debug))

        config_pxi += 'DEF DEBUG = {0}\n'.format(debug)
        config_py += 'DEBUG = {0}\n'.format(debug)
        config_pxi += 'DEF PLATFORM = "{0}"\n'.format(platform)
        config_py += 'PLATFORM = "{0}"\n'.format(platform)
        for fn, content in (
                (config_h_fn, config_h), (config_pxi_fn, config_pxi),
                (config_py_fn, config_py)):
            build_fn = expand(build_path, *fn)
            if self.update_if_changed(build_fn, content):
                print('Updated {}'.format(build_fn))
            src_fn = expand(src_path, *fn)
            if src_fn != build_fn and self.update_if_changed(src_fn, content):
                print('Updated {}'.format(src_fn))

        c = self.compiler.compiler_type
        print('Detected compiler is {}'.format(c))
        if c != 'msvc':
            for e in self.extensions:
                e.extra_link_args += ['-lm']

        super().build_extensions()

    def update_if_changed(self, fn, content):
        need_update = True
        if exists(fn):
            with open(fn) as fd:
                need_update = fd.read() != content
        if need_update:
            directory_name = dirname(fn)
            if not exists(directory_name):
                makedirs(directory_name)
            with open(fn, 'w') as fd:
                fd.write(content)
        return need_update


def _check_and_fix_sdl2_mixer(f_path):
    # Between SDL_mixer 2.0.1 and 2.0.4, the included frameworks changed
    # smpeg2 have been replaced with mpg123, but there is no need to fix.
    smpeg2_path = ("{}/Versions/A/Frameworks/smpeg2.framework"
                   "/Versions/A/smpeg2").format(f_path)
    if not exists(smpeg2_path):
        return

    print("Check if SDL2_mixer smpeg2 have an @executable_path")
    rpath_from = ("@executable_path/../Frameworks/SDL2.framework"
                  "/Versions/A/SDL2")
    rpath_to = "@rpath/../../../../SDL2.framework/Versions/A/SDL2"
    output = getoutput(("otool -L '{}'").format(smpeg2_path)).decode('utf-8')
    if "@executable_path" not in output:
        return

    print("WARNING: Your SDL2_mixer version is invalid")
    print("WARNING: The smpeg2 framework embedded in SDL2_mixer contains a")
    print("WARNING: reference to @executable_path that will fail the")
    print("WARNING: execution of your application.")
    print("WARNING: We are going to change:")
    print("WARNING: from: {}".format(rpath_from))
    print("WARNING: to: {}".format(rpath_to))
    getoutput("install_name_tool -change {} {} {}".format(
        rpath_from, rpath_to, smpeg2_path))

    output = getoutput(("otool -L '{}'").format(smpeg2_path))
    if b"@executable_path" not in output:
        print("WARNING: Change successfully applied!")
        print("WARNING: You'll never see this message again.")
    else:
        print("WARNING: Unable to apply the changes, sorry.")


# -----------------------------------------------------------------------------
print("Python path is:\n{}\n".format('\n'.join(sys.path)))
# extract version (simulate doc generation, kivy will be not imported)
environ['KIVY_DOC_INCLUDE'] = '1'
import kivy

# Cython check
# on python-for-android and kivy-ios, cython usage is external
from kivy.tools.packaging.cython_cfg import get_cython_versions, get_cython_msg
CYTHON_REQUIRES_STRING, MIN_CYTHON_STRING, MAX_CYTHON_STRING, \
    CYTHON_UNSUPPORTED = get_cython_versions()
cython_min_msg, cython_max_msg, cython_unsupported_msg = get_cython_msg()

if can_use_cython:
    import Cython
    from packaging import version
    print('\nFound Cython at', Cython.__file__)

    cy_version_str = Cython.__version__
    cy_ver = version.parse(cy_version_str)
    print('Detected supported Cython version {}'.format(cy_version_str))

    if cy_ver < version.Version(MIN_CYTHON_STRING):
        print(cython_min_msg)
    elif cy_ver in CYTHON_UNSUPPORTED:
        print(cython_unsupported_msg)
    elif cy_ver > version.Version(MAX_CYTHON_STRING):
        print(cython_max_msg)
    sleep(1)

# extra build commands go in the cmdclass dict {'command-name': CommandClass}
# see tools.packaging.{platform}.build.py for custom build commands for
# portable packages. Also e.g. we use build_ext command from cython if its
# installed for c extensions.
from kivy.tools.packaging.factory import FactoryBuild
cmdclass = {
    'build_factory': FactoryBuild,
    'build_ext': KivyBuildExt}

try:
    # add build rules for portable packages to cmdclass
    if platform == 'win32':
        from kivy.tools.packaging.win32.build import WindowsPortableBuild
        cmdclass['build_portable'] = WindowsPortableBuild
    elif platform == 'darwin':
        from kivy.tools.packaging.osx.build import OSXPortableBuild
        cmdclass['build_portable'] = OSXPortableBuild
except ImportError:
    print('User distribution detected, avoid portable command.')

# Detect which opengl version headers to use
if platform in ('android', 'darwin', 'ios', 'rpi', 'mali', 'vc'):
    c_options['use_opengl_es2'] = True
elif c_options['use_opengl_es2'] is None:
    c_options['use_opengl_es2'] = \
        environ.get('KIVY_GRAPHICS', '').lower() == 'gles'

print('Using this graphics system: {}'.format(
    ['OpenGL', 'OpenGL ES 2'][int(c_options['use_opengl_es2'] or False)]))

# check if we are in a kivy-ios build
if platform == 'ios':
    print('Kivy-IOS project environment detect, use it.')
    print('Kivy-IOS project located at {0}'.format(kivy_ios_root))
    c_options['use_ios'] = True
    c_options['use_sdl2'] = True

elif platform == 'android':
    c_options['use_android'] = True

# detect gstreamer, only on desktop
# works if we forced the options or in autodetection
if platform not in ('ios', 'android') and (c_options['use_gstreamer']
                                           in (None, True)):
    gstreamer_valid = False
    if c_options['use_osx_frameworks'] and platform == 'darwin':
        # check the existence of frameworks
        f_path = '/Library/Frameworks/GStreamer.framework'
        if not exists(f_path):
            c_options['use_gstreamer'] = False
            print('GStreamer framework not found, fallback on pkg-config')
        else:
            print('GStreamer framework found')
            gstreamer_valid = True
            c_options['use_gstreamer'] = True
            gst_flags = {
                'extra_link_args': [
                    '-F/Library/Frameworks',
                    '-Xlinker', '-rpath',
                    '-Xlinker', '/Library/Frameworks',
                    '-Xlinker', '-headerpad',
                    '-Xlinker', '190',
                    '-framework', 'GStreamer'],
                'include_dirs': [join(f_path, 'Headers')]}
    elif platform == 'win32':
        gst_flags = pkgconfig('gstreamer-1.0')
        if 'libraries' in gst_flags:
            print('GStreamer found via pkg-config')
            gstreamer_valid = True
            c_options['use_gstreamer'] = True
        else:
            _includes = get_isolated_env_paths()[0] + [
                sysconfig.get_path("include")
            ]
            for include_dir in _includes:
                if exists(join(include_dir, 'gst', 'gst.h')):
                    print('GStreamer found via gst.h')
                    gstreamer_valid = True
                    c_options['use_gstreamer'] = True
                    gst_flags = {
                        'libraries':
                            ['gstreamer-1.0', 'glib-2.0', 'gobject-2.0']}
                    break

    if not gstreamer_valid:
        # use pkg-config approach instead
        gst_flags = pkgconfig('gstreamer-1.0')
        if 'libraries' in gst_flags:
            print('GStreamer found via pkg-config')
            c_options['use_gstreamer'] = True


# detect SDL2, only on desktop and iOS, or android if explicitly enabled
# works if we forced the options or in autodetection
sdl2_flags = {}
sdl2_source = None
if platform == 'win32' and c_options['use_sdl2'] is None:
    c_options['use_sdl2'] = True

can_autodetect_sdl2 = (
    platform not in ("android",) and c_options["use_sdl2"] is None
)
if c_options['use_sdl2'] or can_autodetect_sdl2:

    sdl2_valid = False
    if c_options['use_osx_frameworks'] and platform == 'darwin':
        # check the existence of frameworks
        if KIVY_DEPS_ROOT:
            default_sdl2_frameworks_search_path = join(
                KIVY_DEPS_ROOT, "dist", "Frameworks"
            )
        else:
            default_sdl2_frameworks_search_path = "/Library/Frameworks"
        sdl2_frameworks_search_path = environ.get(
            "KIVY_SDL2_FRAMEWORKS_SEARCH_PATH",
            default_sdl2_frameworks_search_path
        )
        sdl2_valid = True

        sdl2_flags = {
            'extra_link_args': [
                '-F{}'.format(sdl2_frameworks_search_path),
                '-Xlinker', '-rpath',
                '-Xlinker', sdl2_frameworks_search_path,
                '-Xlinker', '-headerpad',
                '-Xlinker', '190'],
            'include_dirs': [],
            'extra_compile_args': ['-F{}'.format(sdl2_frameworks_search_path)]
        }

        for name in ('SDL2', 'SDL2_ttf', 'SDL2_image', 'SDL2_mixer'):
            f_path = '{}/{}.framework'.format(sdl2_frameworks_search_path, name)
            if not exists(f_path):
                print('Missing framework {}'.format(f_path))
                sdl2_valid = False
                continue
            sdl2_flags['extra_link_args'] += ['-framework', name]
            sdl2_flags['include_dirs'] += [join(f_path, 'Headers')]
            print('Found sdl2 frameworks: {}'.format(f_path))
            if name == 'SDL2_mixer':
                _check_and_fix_sdl2_mixer(f_path)

        if not sdl2_valid:
            c_options['use_sdl2'] = False
            print('SDL2 frameworks not found, fallback on pkg-config')
        else:
            c_options['use_sdl2'] = True
            sdl2_source = 'macos-frameworks'
            print('Activate SDL2 compilation')

    if not sdl2_valid and platform != "ios":
        # use pkg-config approach instead
        sdl2_flags = pkgconfig('sdl2', 'SDL2_ttf', 'SDL2_image', 'SDL2_mixer')
        if 'libraries' in sdl2_flags:
            print('SDL2 found via pkg-config')
            c_options['use_sdl2'] = True
            sdl2_source = 'pkg-config'


# -----------------------------------------------------------------------------
# declare flags

def expand(root, *args):
    return join(root, 'kivy', *args)


class CythonExtension(Extension):

    def __init__(self, *args, **kwargs):
        Extension.__init__(self, *args, **kwargs)
        self.cython_directives = {
            'c_string_encoding': 'utf-8',
            'profile': 'USE_PROFILE' in environ,
            'embedsignature': use_embed_signature,
            'language_level': 3,
            'unraisable_tracebacks': True,
        }


def merge(d1, *args):
    d1 = deepcopy(d1)
    for d2 in args:
        for key, value in d2.items():
            value = deepcopy(value)
            if key in d1:
                d1[key].extend(value)
            else:
                d1[key] = value
    return d1


def determine_base_flags():
    includes, libs = get_isolated_env_paths()

    flags = {
        'libraries': [],
        'include_dirs': [join(src_path, 'kivy', 'include')] + includes,
        'library_dirs': [] + libs,
        'extra_link_args': [],
        'extra_compile_args': []}
    if c_options['use_ios']:
        sysroot = environ.get('IOSSDKROOT', environ.get('SDKROOT'))
        if not sysroot:
            raise Exception('IOSSDKROOT is not set')
        flags['include_dirs'] += [sysroot]
        flags['extra_compile_args'] += ['-isysroot', sysroot]
        flags['extra_link_args'] += ['-isysroot', sysroot]
    elif platform.startswith('freebsd'):
        flags['include_dirs'] += [join(
            environ.get('LOCALBASE', '/usr/local'), 'include')]
        flags['library_dirs'] += [join(
            environ.get('LOCALBASE', '/usr/local'), 'lib')]
    elif platform == 'darwin' and c_options['use_osx_frameworks']:
        v = os.uname()
        if v[2] >= '13.0.0':
            if 'SDKROOT' in environ:
                sysroot = join(environ['SDKROOT'], 'System/Library/Frameworks')
            else:
                # use xcode-select to search on the right Xcode path
                # XXX use the best SDK available instead of a specific one
                import platform as _platform
                xcode_dev = getoutput('xcode-select -p').splitlines()[0]
                sdk_mac_ver = '.'.join(_platform.mac_ver()[0].split('.')[:2])
                print('Xcode detected at {}, and using OS X{} sdk'.format(
                    xcode_dev, sdk_mac_ver))
                sysroot = join(
                    xcode_dev.decode('utf-8'),
                    'Platforms/MacOSX.platform/Developer/SDKs',
                    'MacOSX{}.sdk'.format(sdk_mac_ver),
                    'System/Library/Frameworks')
        else:
            sysroot = ('/System/Library/Frameworks/'
                       'ApplicationServices.framework/Frameworks')
        flags['extra_compile_args'] += ['-F%s' % sysroot]
        flags['extra_link_args'] += ['-F%s' % sysroot]
    elif platform == 'win32':
        flags['include_dirs'] += [sysconfig.get_path('include')]
        flags['library_dirs'] += [join(sys.prefix, "libs")]
    return flags


def determine_gl_flags():
    kivy_graphics_include = join(src_path, 'kivy', 'include')
    flags = {'include_dirs': [kivy_graphics_include], 'libraries': []}
    base_flags = {'include_dirs': [kivy_graphics_include], 'libraries': []}
    cross_sysroot = environ.get('KIVY_CROSS_SYSROOT')

    if c_options['use_opengl_mock']:
        return flags, base_flags
    if platform == 'win32':
        flags['libraries'] = ['opengl32', 'glew32']
    elif platform == 'ios':
        flags['libraries'] = ['GLESv2']
        flags['extra_link_args'] = ['-framework', 'OpenGLES']
    elif platform == 'darwin':
        flags['extra_link_args'] = ['-framework', 'OpenGL']
    elif platform.startswith('freebsd'):
        flags['libraries'] = ['GL']
    elif platform.startswith('openbsd'):
        flags['include_dirs'] = ['/usr/X11R6/include']
        flags['library_dirs'] = ['/usr/X11R6/lib']
        flags['libraries'] = ['GL']
    elif platform == 'android':
        flags['include_dirs'] = [join(ndkplatform, 'usr', 'include')]
        flags['library_dirs'] = [join(ndkplatform, 'usr', 'lib')]
        flags['libraries'] = ['GLESv2']
    elif platform == 'rpi':

        if not cross_sysroot:
            flags['include_dirs'] = [
                '/opt/vc/include',
                '/opt/vc/include/interface/vcos/pthreads',
                '/opt/vc/include/interface/vmcs_host/linux']
            flags['library_dirs'] = ['/opt/vc/lib']
            brcm_lib_files = (
                '/opt/vc/lib/libbrcmEGL.so',
                '/opt/vc/lib/libbrcmGLESv2.so')

        else:
            print("KIVY_CROSS_SYSROOT: " + cross_sysroot)
            flags['include_dirs'] = [
                cross_sysroot + '/usr/include',
                cross_sysroot + '/usr/include/interface/vcos/pthreads',
                cross_sysroot + '/usr/include/interface/vmcs_host/linux']
            flags['library_dirs'] = [cross_sysroot + '/usr/lib']
            brcm_lib_files = (
                cross_sysroot + '/usr/lib/libbrcmEGL.so',
                cross_sysroot + '/usr/lib/libbrcmGLESv2.so')

        if all((exists(lib) for lib in brcm_lib_files)):
            print('Found brcmEGL and brcmGLES library files '
                  'for rpi platform at ' + dirname(brcm_lib_files[0]))
            gl_libs = ['brcmEGL', 'brcmGLESv2']
        else:
            print(
                'Failed to find brcmEGL and brcmGLESv2 library files '
                'for rpi platform, falling back to EGL and GLESv2.')
            gl_libs = ['EGL', 'GLESv2']
        flags['libraries'] = ['bcm_host'] + gl_libs
    elif platform in ['mali', 'vc']:
        flags['include_dirs'] = ['/usr/include/']
        flags['library_dirs'] = ['/usr/lib/arm-linux-gnueabihf']
        flags['libraries'] = ['GLESv2']
        c_options['use_x11'] = True
        c_options['use_egl'] = True
    else:
        flags['libraries'] = ['GL']
    return flags, base_flags


def determine_sdl2():
    flags = {}
    if not c_options['use_sdl2']:
        return flags

    # If darwin has already been configured with frameworks, don't
    # configure sdl2 via libs.
    # TODO: Move framework configuration here.
    if sdl2_source == "macos-frameworks":
        return sdl2_flags

    default_sdl2_path = None

    if KIVY_DEPS_ROOT:

        default_sdl2_path = os.pathsep.join(
            [
                join(KIVY_DEPS_ROOT, "dist", "lib"),
                join(KIVY_DEPS_ROOT, "dist", "lib64"),
                join(KIVY_DEPS_ROOT, "dist", "include", "SDL2"),
            ]
        )

    kivy_sdl2_path = environ.get('KIVY_SDL2_PATH', default_sdl2_path)

    includes, _ = get_isolated_env_paths()

    # no pkgconfig info, or we want to use a specific sdl2 path, so perform
    # manual configuration
    flags['libraries'] = ['SDL2', 'SDL2_ttf', 'SDL2_image', 'SDL2_mixer']

    sdl2_paths = kivy_sdl2_path.split(os.pathsep) if kivy_sdl2_path else []

    if not sdl2_paths:
        # Try to find sdl2 in default locations if we don't have a custom path
        sdl2_paths = []
        for include in includes + [join(sys.prefix, 'include')]:
            sdl_inc = join(include, 'SDL2')
            if isdir(sdl_inc):
                sdl2_paths.append(sdl_inc)
        sdl2_paths.extend(['/usr/local/include/SDL2', '/usr/include/SDL2'])

    flags['include_dirs'] = sdl2_paths
    flags['extra_link_args'] = []
    flags['extra_compile_args'] = []
    flags['library_dirs'] = (
        sdl2_paths if sdl2_paths else
        ['/usr/local/lib/'])

    if kivy_sdl2_path:
        # If we have a custom path, we need to add the rpath to the linker
        # so that the libraries can be found and loaded without having to
        # set LD_LIBRARY_PATH every time.
        flags["extra_link_args"] = [
            f"-Wl,-rpath,{l_path}"
            for l_path in sdl2_paths
            if l_path.endswith("lib")
        ]

    if sdl2_flags:
        flags = merge(flags, sdl2_flags)

    # ensure headers for all the SDL2 and sub libraries are available
    libs_to_check = ['SDL', 'SDL_mixer', 'SDL_ttf', 'SDL_image']
    can_compile = True
    for lib in libs_to_check:
        found = False
        for d in flags['include_dirs']:
            fn = join(d, '{}.h'.format(lib))
            if exists(fn):
                found = True
                print('SDL2: found {} header at {}'.format(lib, fn))
                break

        if not found:
            print('SDL2: missing sub library {}'.format(lib))
            can_compile = False

    if not can_compile:
        c_options['use_sdl2'] = False
        return {}

    return flags


base_flags = determine_base_flags()
gl_flags, gl_flags_base = determine_gl_flags()

# -----------------------------------------------------------------------------
# sources to compile
# all the dependencies have been found manually with:
# grep -inr -E '(cimport|include)' kivy/graphics/context_instructions.{pxd,pyx}
graphics_dependencies = {
    'boxshadow.pxd': ['fbo.pxd', 'context_instructions.pxd',
                      'vertex_instructions.pxd', 'instructions.pxd'],
    'boxshadow.pyx': ['fbo.pxd', 'context_instructions.pxd',
                      'instructions.pyx'],
    'buffer.pyx': ['common.pxi'],
    'context.pxd': ['instructions.pxd', 'texture.pxd', 'vbo.pxd', 'cgl.pxd'],
    'cgl.pxd': ['common.pxi', 'config.pxi', 'gl_redirect.h'],
    'compiler.pxd': ['instructions.pxd'],
    'compiler.pyx': ['context_instructions.pxd'],
    'cgl.pyx': ['cgl.pxd'],
    'cgl_mock.pyx': ['cgl.pxd'],
    'cgl_sdl2.pyx': ['cgl.pxd'],
    'cgl_gl.pyx': ['cgl.pxd'],
    'cgl_glew.pyx': ['cgl.pxd'],
    'context_instructions.pxd': [
        'transformation.pxd', 'instructions.pxd', 'texture.pxd'],
    'fbo.pxd': ['cgl.pxd', 'instructions.pxd', 'texture.pxd'],
    'fbo.pyx': [
        'config.pxi', 'opcodes.pxi', 'transformation.pxd', 'context.pxd'],
    'gl_instructions.pyx': [
        'config.pxi', 'opcodes.pxi', 'cgl.pxd', 'instructions.pxd'],
    'instructions.pxd': [
        'vbo.pxd', 'context_instructions.pxd', 'compiler.pxd', 'shader.pxd',
        'texture.pxd', '../_event.pxd'],
    'instructions.pyx': [
        'config.pxi', 'opcodes.pxi', 'cgl.pxd',
        'context.pxd', 'common.pxi', 'vertex.pxd', 'transformation.pxd'],
    'opengl.pyx': [
        'config.pxi', 'common.pxi', 'cgl.pxd', 'gl_redirect.h'],
    'opengl_utils.pyx': [
        'opengl_utils_def.pxi', 'cgl.pxd', ],
    'shader.pxd': ['cgl.pxd', 'transformation.pxd', 'vertex.pxd'],
    'shader.pyx': [
        'config.pxi', 'common.pxi', 'cgl.pxd',
        'vertex.pxd', 'transformation.pxd', 'context.pxd',
        'gl_debug_logger.pxi'],
    'stencil_instructions.pxd': ['instructions.pxd'],
    'stencil_instructions.pyx': [
        'config.pxi', 'opcodes.pxi', 'cgl.pxd',
        'gl_debug_logger.pxi'],
    'scissor_instructions.pyx': [
        'config.pxi', 'opcodes.pxi', 'cgl.pxd'],
    'svg.pyx': ['config.pxi', 'common.pxi', 'texture.pxd', 'instructions.pxd',
                'vertex_instructions.pxd', 'tesselator.pxd'],
    'texture.pxd': ['cgl.pxd'],
    'texture.pyx': [
        'config.pxi', 'common.pxi', 'opengl_utils_def.pxi', 'context.pxd',
        'cgl.pxd', 'opengl_utils.pxd',
        'img_tools.pxi', 'gl_debug_logger.pxi'],
    'vbo.pxd': ['buffer.pxd', 'cgl.pxd', 'vertex.pxd'],
    'vbo.pyx': [
        'config.pxi', 'common.pxi', 'context.pxd',
        'instructions.pxd', 'shader.pxd', 'gl_debug_logger.pxi'],
    'vertex.pxd': ['cgl.pxd'],
    'vertex.pyx': ['config.pxi', 'common.pxi'],
    'vertex_instructions.pyx': [
        'config.pxi', 'common.pxi', 'vbo.pxd', 'vertex.pxd',
        'instructions.pxd', 'vertex_instructions.pxd',
        'cgl.pxd', 'texture.pxd', 'vertex_instructions_line.pxi'],
    'vertex_instructions_line.pxi': ['stencil_instructions.pxd']}

sources = {
    '_event.pyx': merge(base_flags, {'depends': ['properties.pxd']}),
    '_clock.pyx': {},
    'weakproxy.pyx': {},
    'properties.pyx': merge(
        base_flags, {'depends': ['_event.pxd', '_metrics.pxd']}),
    '_metrics.pyx': merge(base_flags, {'depends': ['_event.pxd']}),
    'graphics/buffer.pyx': merge(base_flags, gl_flags_base),
    'graphics/context.pyx': merge(base_flags, gl_flags_base),
    'graphics/compiler.pyx': merge(base_flags, gl_flags_base),
    'graphics/context_instructions.pyx': merge(base_flags, gl_flags_base),
    'graphics/fbo.pyx': merge(base_flags, gl_flags_base),
    'graphics/gl_instructions.pyx': merge(base_flags, gl_flags_base),
    'graphics/instructions.pyx': merge(base_flags, gl_flags_base),
    'graphics/opengl.pyx': merge(base_flags, gl_flags_base),
    'graphics/opengl_utils.pyx': merge(base_flags, gl_flags_base),
    'graphics/shader.pyx': merge(base_flags, gl_flags_base),
    'graphics/stencil_instructions.pyx': merge(base_flags, gl_flags_base),
    'graphics/scissor_instructions.pyx': merge(base_flags, gl_flags_base),
    'graphics/texture.pyx': merge(base_flags, gl_flags_base),
    'graphics/transformation.pyx': merge(base_flags, gl_flags_base),
    'graphics/vbo.pyx': merge(base_flags, gl_flags_base),
    'graphics/vertex.pyx': merge(base_flags, gl_flags_base),
    'graphics/vertex_instructions.pyx': merge(base_flags, gl_flags_base),
    'graphics/cgl.pyx': merge(base_flags, gl_flags_base),
    'graphics/cgl_backend/cgl_mock.pyx': merge(base_flags, gl_flags_base),
    'graphics/cgl_backend/cgl_gl.pyx': merge(base_flags, gl_flags),
    'graphics/cgl_backend/cgl_glew.pyx': merge(base_flags, gl_flags),
    'graphics/cgl_backend/cgl_sdl2.pyx': merge(base_flags, gl_flags_base),
    'graphics/cgl_backend/cgl_debug.pyx': merge(base_flags, gl_flags_base),
    'core/text/text_layout.pyx': base_flags,
    'core/window/window_info.pyx': base_flags,
    'graphics/tesselator.pyx': merge(base_flags, {
        'include_dirs': ['kivy/lib/libtess2/Include'],
        'c_depends': [
            'lib/libtess2/Source/bucketalloc.c',
            'lib/libtess2/Source/dict.c',
            'lib/libtess2/Source/geom.c',
            'lib/libtess2/Source/mesh.c',
            'lib/libtess2/Source/priorityq.c',
            'lib/libtess2/Source/sweep.c',
            'lib/libtess2/Source/tess.c'
        ]
    }),
    'graphics/svg.pyx': merge(base_flags, gl_flags_base),
    'graphics/boxshadow.pyx': merge(base_flags, gl_flags_base)
}

if c_options["use_sdl2"]:
    sdl2_flags = determine_sdl2()

if c_options['use_sdl2'] and sdl2_flags:
    sources['graphics/cgl_backend/cgl_sdl2.pyx'] = merge(
        sources['graphics/cgl_backend/cgl_sdl2.pyx'], sdl2_flags)
    sdl2_depends = {'depends': ['lib/sdl2.pxi']}
    for source_file in ('core/window/_window_sdl2.pyx',
                        'core/image/_img_sdl2.pyx',
                        'core/text/_text_sdl2.pyx',
                        'core/audio/audio_sdl2.pyx',
                        'core/clipboard/_clipboard_sdl2.pyx'):
        sources[source_file] = merge(
            base_flags, sdl2_flags, sdl2_depends)

if c_options['use_pangoft2'] in (None, True) and platform not in (
                                      'android', 'ios', 'win32'):
    pango_flags = pkgconfig('pangoft2')
    if pango_flags and 'libraries' in pango_flags:
        print('Pango: pangoft2 found via pkg-config')
        c_options['use_pangoft2'] = True
        pango_depends = {'depends': [
            'lib/pango/pangoft2.pxi',
            'lib/pango/pangoft2.h']}
        sources['core/text/_text_pango.pyx'] = merge(
                base_flags, pango_flags, pango_depends)
        print(sources['core/text/_text_pango.pyx'])

if platform in ('darwin', 'ios'):
    # activate ImageIO provider for our core image
    if platform == 'ios':
        osx_flags = {'extra_link_args': [
            '-framework', 'Foundation',
            '-framework', 'UIKit',
            '-framework', 'AudioToolbox',
            '-framework', 'CoreGraphics',
            '-framework', 'QuartzCore',
            '-framework', 'ImageIO',
            '-framework', 'Accelerate']}
    else:
        osx_flags = {'extra_link_args': [
            '-framework', 'ApplicationServices']}
    sources['core/image/img_imageio.pyx'] = merge(
        base_flags, osx_flags)

if c_options['use_avfoundation']:
    import platform as _platform
    mac_ver = [int(x) for x in _platform.mac_ver()[0].split('.')[:2]]
    if mac_ver >= [10, 7] or platform == 'ios':
        osx_flags = {
            'extra_link_args': ['-framework', 'AVFoundation'],
            'extra_compile_args': ['-ObjC++'],
            'depends': ['core/camera/camera_avfoundation_implem.m']}
        sources['core/camera/camera_avfoundation.pyx'] = merge(
            base_flags, osx_flags)
    else:
        print('AVFoundation cannot be used, OSX >= 10.7 is required')

if c_options['use_rpi_vidcore_lite']:

    # DISPMANX is only available on old versions of Raspbian (Buster).
    # For this reason, we need to be sure that EGL_DISPMANX_* is available
    # before compiling the vidcore_lite module, even if we're on a RPi.
    HAVE_DISPMANX = check_c_source_compiles(
        textwrap.dedent(
            """
        #include <bcm_host.h>
        #include <EGL/eglplatform.h>
        int main(int argc, char **argv) {
            EGL_DISPMANX_WINDOW_T window;
            bcm_host_init();
        }
        """
        ),
        include_dirs=gl_flags["include_dirs"],
    )
    if HAVE_DISPMANX:
        sources['lib/vidcore_lite/egl.pyx'] = merge(
            base_flags, gl_flags)
        sources['lib/vidcore_lite/bcm.pyx'] = merge(
            base_flags, gl_flags)

if c_options['use_x11']:
    libs = ['Xrender', 'X11']
    if c_options['use_egl']:
        libs += ['EGL']
    else:
        libs += ['GL']
    sources['core/window/window_x11.pyx'] = merge(
        base_flags, gl_flags, {
            # FIXME add an option to depend on them but not compile them
            # cause keytab is included in core, and core is included in
            # window_x11
            #
            # 'depends': [
            #    'core/window/window_x11_keytab.c',
            #    'core/window/window_x11_core.c'],
            'libraries': libs})

if c_options['use_gstreamer']:
    sources['lib/gstplayer/_gstplayer.pyx'] = merge(
        base_flags, gst_flags, {
            'depends': ['lib/gstplayer/_gstplayer.h']})

# -----------------------------------------------------------------------------
# extension modules


def get_dependencies(name, deps=None):
    if deps is None:
        deps = []
    for dep in graphics_dependencies.get(name, []):
        if dep not in deps:
            deps.append(dep)
            get_dependencies(dep, deps)
    return deps


def resolve_dependencies(fn, depends):
    fn = basename(fn)
    deps = []
    get_dependencies(fn, deps)
    get_dependencies(fn.replace('.pyx', '.pxd'), deps)

    deps_final = []
    paths_to_test = ['graphics', 'include']
    for dep in deps:
        found = False
        for path in paths_to_test:
            filename = expand(src_path, path, dep)
            if exists(filename):
                deps_final.append(filename)
                found = True
                break
        if not found:
            print('ERROR: Dependency for {} not resolved: {}'.format(
                fn, dep
            ))

    return deps_final


def get_extensions_from_sources(sources):
    ext_modules = []
    if environ.get('KIVY_FAKE_BUILDEXT'):
        print('Fake build_ext asked, will generate only .h/.c')
        return ext_modules
    for pyx, flags in sources.items():
        is_graphics = pyx.startswith('graphics')
        pyx_path = expand(src_path, pyx)
        depends = [expand(src_path, x) for x in flags.pop('depends', [])]
        c_depends = [expand(src_path, x) for x in flags.pop('c_depends', [])]
        if not can_use_cython:
            # can't use cython, so use the .c files instead.
            pyx_path = '%s.c' % pyx_path[:-4]
        if is_graphics:
            depends = resolve_dependencies(pyx_path, depends)
        f_depends = [x for x in depends if x.rsplit('.', 1)[-1] in (
            'c', 'cpp', 'm')]
        module_name = '.'.join(['kivy'] + pyx[:-4].split('/'))
        flags_clean = {'depends': depends}
        for key, value in flags.items():
            if len(value):
                flags_clean[key] = value
        ext_modules.append(CythonExtension(
            module_name, [pyx_path] + f_depends + c_depends, **flags_clean))
    return ext_modules


ext_modules = get_extensions_from_sources(sources)


# -----------------------------------------------------------------------------
# automatically detect data files
split_examples = int(environ.get('KIVY_SPLIT_EXAMPLES', '0'))
data_file_prefix = 'share/kivy-'
examples = {}
examples_allowed_ext = ('readme', 'py', 'wav', 'png', 'jpg', 'svg', 'json',
                        'avi', 'gif', 'txt', 'ttf', 'obj', 'mtl', 'kv', 'mpg',
                        'glsl', 'zip')
for root, subFolders, files in walk('examples'):
    for fn in files:
        ext = fn.split('.')[-1].lower()
        if ext not in examples_allowed_ext:
            continue
        filename = join(root, fn)
        directory = '%s%s' % (data_file_prefix, dirname(filename))
        if directory not in examples:
            examples[directory] = []
        examples[directory].append(filename)

binary_deps = []
binary_deps_path = join(src_path, 'kivy', 'binary_deps')
if isdir(binary_deps_path):
    for root, dirnames, filenames in walk(binary_deps_path):
        for fname in filenames:
            binary_deps.append(
                join(root.replace(binary_deps_path, 'binary_deps'), fname))


def glob_paths(*patterns, excludes=('.pyc', )):
    files = []
    base = Path(join(src_path, 'kivy'))

    for pat in patterns:
        for f in base.glob(pat):
            if f.suffix in excludes:
                continue
            files.append(str(f.relative_to(base)))
    return files


# -----------------------------------------------------------------------------
# setup !
if not build_examples:
    setup(
        name='Kivy',
        version=__version__,
        author='Kivy Team and other contributors',
        author_email='kivy-dev@googlegroups.com',
        url='http://kivy.org',
        project_urls={
            'Source': 'https://github.com/kivy/kivy',
            'Documentation': 'https://kivy.org/doc/stable/',
            'Bug Reports': "https://github.com/kivy/kivy/issues",
        },
        license='MIT',
        description=(
            'An open-source Python framework for developing '
            'GUI apps that work cross-platform, including '
            'desktop, mobile and embedded platforms.'),
        long_description=get_description(),
        long_description_content_type='text/markdown',
        ext_modules=ext_modules,
        cmdclass=cmdclass,
        packages=find_packages(include=['kivy*']),
        package_dir={'kivy': 'kivy'},
        package_data={
            'kivy':
                glob_paths('*.pxd', '*.pxi') +
                glob_paths('**/*.pxd', '**/*.pxi') +
                glob_paths('data/**/*.*') +
                glob_paths('include/**/*.*') +
                glob_paths('tools/**/*.*', excludes=('.pyc', '.enc')) +
                glob_paths('graphics/**/*.h') +
                glob_paths('tests/**/*.*') +
                [
                    'setupconfig.py',
                ] + binary_deps
        },
        data_files=[] if split_examples else list(examples.items()),
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Environment :: MacOS X',
            'Environment :: Win32 (MS Windows)',
            'Environment :: X11 Applications',
            'Intended Audience :: Developers',
            'Intended Audience :: End Users/Desktop',
            'Intended Audience :: Information Technology',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: MIT License',
            'Natural Language :: English',
            'Operating System :: MacOS :: MacOS X',
            'Operating System :: Microsoft :: Windows',
            'Operating System :: POSIX :: BSD :: FreeBSD',
            'Operating System :: POSIX :: Linux',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            'Programming Language :: Python :: 3.11',
            'Programming Language :: Python :: 3.12',
            'Programming Language :: Python :: 3.13',
            'Topic :: Artistic Software',
            'Topic :: Games/Entertainment',
            'Topic :: Multimedia :: Graphics :: 3D Rendering',
            'Topic :: Multimedia :: Graphics :: Capture :: Digital Camera',
            'Topic :: Multimedia :: Graphics :: Presentation',
            'Topic :: Multimedia :: Graphics :: Viewers',
            'Topic :: Multimedia :: Sound/Audio :: Players :: MP3',
            'Topic :: Multimedia :: Video :: Display',
            'Topic :: Scientific/Engineering :: Human Machine Interfaces',
            'Topic :: Scientific/Engineering :: Visualization',
            ('Topic :: Software Development :: Libraries :: '
             'Application Frameworks'),
            'Topic :: Software Development :: User Interfaces'])
else:
    setup(
        name='Kivy-examples',
        version=__version__,
        author='Kivy Team and other contributors',
        author_email='kivy-dev@googlegroups.com',
        url='http://kivy.org',
        license='MIT',
        description=('Kivy examples.'),
        long_description_content_type='text/markdown',
        long_description=get_description(),
        data_files=list(examples.items()))



'''
Framebuffer
===========

The Fbo is like an offscreen window. You can activate the fbo for rendering into
a texture and use your fbo as a texture for other drawing.

The Fbo acts as a :class:`kivy.graphics.instructions.Canvas`.

Here is an example of using an fbo for some colored rectangles::

    from kivy.graphics import Fbo, Color, Rectangle

    class FboTest(Widget):
        def __init__(self, **kwargs):
            super(FboTest, self).__init__(**kwargs)

            # first step is to create the fbo and use the fbo texture on other
            # rectangle

            with self.canvas:
                # create the fbo
                self.fbo = Fbo(size=(256, 256))

                # show our fbo on the widget in different size
                Color(1, 1, 1)
                Rectangle(size=(32, 32), texture=self.fbo.texture)
                Rectangle(pos=(32, 0), size=(64, 64), texture=self.fbo.texture)
                Rectangle(pos=(96, 0), size=(128, 128), texture=self.fbo.texture)

            # in the second step, you can draw whatever you want on the fbo
            with self.fbo:
                Color(1, 0, 0, .8)
                Rectangle(size=(256, 64))
                Color(0, 1, 0, .8)
                Rectangle(size=(64, 256))

If you change anything in the `self.fbo` object, it will be automatically updated.
The canvas where the fbo is put will be automatically updated as well.

Reloading the FBO content
-------------------------

.. versionadded:: 1.2.0

If the OpenGL context is lost, then the FBO is lost too. You need to reupload
data on it yourself. Use the :meth:`Fbo.add_reload_observer` to add a reloading
function that will be automatically called when needed::

    def __init__(self, **kwargs):
        super(...).__init__(**kwargs)
        self.fbo = Fbo(size=(512, 512))
        self.fbo.add_reload_observer(self.populate_fbo)

        # and load the data now.
        self.populate_fbo(self.fbo)


    def populate_fbo(self, fbo):
        with fbo:
            # .. put your Color / Rectangle / ... here

This way, you could use the same method for initialization and for reloading.
But it's up to you.

'''

__all__ = ('Fbo', )

include "../include/config.pxi"
include "opcodes.pxi"

from os import environ
from kivy.compat import PY2
from kivy.logger import Logger
from kivy.utils import platform
from kivy.weakmethod import WeakMethod
from kivy.graphics.texture cimport Texture
from kivy.graphics.transformation cimport Matrix
from kivy.graphics.context cimport get_context

from kivy.graphics.cgl cimport *

from kivy.graphics.instructions cimport RenderContext, Canvas
from kivy.graphics.opengl import glReadPixels as py_glReadPixels
from kivy.graphics.stencil_instructions cimport (
    get_stencil_state, restore_stencil_state, reset_stencil_state)

cdef list fbo_stack = []
cdef list fbo_release_list = []

# must be running GLES2 not just the headers
cdef int IS_GLES_PLATFORM = USE_OPENGL_ES2
if platform == "darwin":
    IS_GLES_PLATFORM = 0


# XXX OpenGL & GLES 3 symbol, but works and required on "GLES 2" on iOS
# Adding it to headers can be complicated because it doesn't exists in gl2.h
cdef int GL_DEPTH_STENCIL_ATTACHMENT = 0x821A


cdef class Fbo(RenderContext):
    '''Fbo class for wrapping the OpenGL Framebuffer extension. The Fbo support
    "with" statement.

    :Parameters:
        `clear_color`: tuple, defaults to (0, 0, 0, 0)
            Define the default color for clearing the framebuffer
        `size`: tuple, defaults to (1024, 1024)
            Default size of the framebuffer
        `push_viewport`: bool, defaults to True
            If True, the OpenGL viewport will be set to the framebuffer size,
            and will be automatically restored when the framebuffer released.
        `with_depthbuffer`: bool, defaults to False
            If True, the framebuffer will be allocated with a Z buffer.
        `with_stencilbuffer`: bool, defaults to False
            .. versionadded:: 1.9.0

            If True, the framebuffer will be allocated with a stencil buffer.
        `texture`: :class:`~kivy.graphics.texture.Texture`, defaults to None
            If None, a default texture will be created.

    .. note::
        Using both of ``with_stencilbuffer`` and ``with_depthbuffer`` is not
        supported in kivy 1.9.0

    '''
    cdef str resolve_status(self, int status):
        if status == GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT:
            return 'Incomplete attachment'
        elif status == GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS:
            return 'Incomplete dimensions'
        elif status == GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT:
            return 'Incomplete missing attachment'
        elif status == GL_FRAMEBUFFER_UNSUPPORTED:
            return 'Unsupported'
        elif status == GL_FRAMEBUFFER_UNDEFINED_OES:
            return 'Undefined framebuffer'
        elif status == 0x8219: #GL_FRAMEBUFFER_UNDEFINED
            return 'Undefined framebuffer'
        elif status == 0x8cdb: #GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER
            return 'Incomplete draw buffer'
        elif status == 0x8cdc: #GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER
            return 'Incomplete read buffer'
        elif status == 0x8d56: #GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE
            return 'Incomplete multisample'
        elif status == 0x8da8: #GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS
            return 'Incomplete layer targets'
        elif status == 0x8da9: #GL_FRAMEBUFFER_INCOMPLETE_LAYER_COUNT
            return 'Incomplete layer count'

        return 'Unknown (status=%x)' % status

    cdef void raise_exception(self, str message, int status=0):
        if status:
            message += ': %s (%d)' % (self.resolve_status(status), status)
        raise Exception(message)

    def __init__(self, *args, **kwargs):
        get_context().register_fbo(self)

        RenderContext.__init__(self, *args, **kwargs)

        if 'clear_color' not in kwargs:
            kwargs['clear_color'] = (0, 0, 0, 0)
        if 'size' not in kwargs:
            kwargs['size'] = (1024, 1024)
        if 'push_viewport' not in kwargs:
            kwargs['push_viewport'] = True
        if 'with_depthbuffer' not in kwargs:
            kwargs['with_depthbuffer'] = False
        if 'with_stencilbuffer' not in kwargs:
            kwargs['with_stencilbuffer'] = False
        if 'texture' not in kwargs:
            kwargs['texture'] = None

        self.buffer_id = 0
        self.depthbuffer_id = 0
        self.stencilbuffer_id = 0
        self._width, self._height  = kwargs['size']
        self.clear_color = kwargs['clear_color']
        self._depthbuffer_attached = int(kwargs['with_depthbuffer'])
        self._stencilbuffer_attached = int(kwargs['with_stencilbuffer'])
        self._push_viewport = int(kwargs['push_viewport'])
        self._is_bound = 0
        self._texture = kwargs['texture']
        self.observers = []

        if IS_GLES_PLATFORM:
            # on GLES/iOS, depthbuffer works with stencil enabled
            # that's how openFrameworks did for all GLES platform in ofFbo.cpp
            if self._depthbuffer_attached:
                self._stencilbuffer_attached = True

        self.create_fbo()

    def __dealloc__(self):
        get_context().dealloc_fbo(self)

    cdef void delete_fbo(self):
        self._texture = None
        get_context().dealloc_fbo(self)
        self.buffer_id = 0
        self.depthbuffer_id = 0

    cdef void create_fbo(self):
        cdef GLuint f_id = 0
        cdef GLint old_fid = 0
        cdef GLint old_rid = 0
        cdef int status
        cdef int do_clear = 0

        # create texture
        if self._texture is None:
            self._texture = Texture.create(size=(self._width, self._height))
            do_clear = 1

        # apply any changes if needed
        self._texture.bind()

        # create framebuffer
        cgl.glGenFramebuffers(1, &f_id)
        self.buffer_id = f_id
        cgl.glGetIntegerv(GL_FRAMEBUFFER_BINDING, &old_fid)
        cgl.glBindFramebuffer(GL_FRAMEBUFFER, self.buffer_id)

        if self._depthbuffer_attached or self._stencilbuffer_attached:
            cgl.glGetIntegerv(GL_RENDERBUFFER_BINDING, &old_rid)

        # depth+stencil renderbuffer
        if self._depthbuffer_attached and self._stencilbuffer_attached:
            cgl.glGenRenderbuffers(1, &f_id)
            self.depthbuffer_id = self.stencilbuffer_id = f_id
            cgl.glBindRenderbuffer(GL_RENDERBUFFER, f_id)
            cgl.glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH24_STENCIL8_OES,
                                  self._width, self._height)
            cgl.glBindRenderbuffer(GL_RENDERBUFFER, old_rid)

            if IS_GLES_PLATFORM:
                cgl.glFramebufferRenderbuffer(
                    GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT,
                    GL_RENDERBUFFER, f_id)
            else:
                cgl.glFramebufferRenderbuffer(
                    GL_FRAMEBUFFER, GL_DEPTH_STENCIL_ATTACHMENT,
                    GL_RENDERBUFFER, f_id)

        # if we need depth, create a renderbuffer
        elif self._depthbuffer_attached:
            cgl.glGenRenderbuffers(1, &f_id)
            self.depthbuffer_id = f_id
            cgl.glBindRenderbuffer(GL_RENDERBUFFER, self.depthbuffer_id)
            cgl.glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT16,
                                  self._width, self._height)
            cgl.glBindRenderbuffer(GL_RENDERBUFFER, old_rid)
            cgl.glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT,
                                      GL_RENDERBUFFER, self.depthbuffer_id)

        # if we need stencil, create a renderbuffer
        elif self._stencilbuffer_attached:
            cgl.glGenRenderbuffers(1, &f_id)
            self.stencilbuffer_id = f_id
            cgl.glBindRenderbuffer(GL_RENDERBUFFER, self.stencilbuffer_id)
            cgl.glRenderbufferStorage(GL_RENDERBUFFER, GL_STENCIL_INDEX8,
                                  self._width, self._height)
            cgl.glBindRenderbuffer(GL_RENDERBUFFER, old_rid)
            cgl.glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_STENCIL_ATTACHMENT,
                                      GL_RENDERBUFFER, self.stencilbuffer_id)

        # attach the framebuffer to our texture
        cgl.glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,
                self._texture._target, self._texture._id, 0)

        # check the status of the framebuffer
        status = cgl.glCheckFramebufferStatus(GL_FRAMEBUFFER)

        if (status == GL_FRAMEBUFFER_UNSUPPORTED and
                (self._stencilbuffer_attached ^ self._depthbuffer_attached)):
            # attempt to automatically fall back to a depth+stencil buffer
            Logger.warning('Fbo: unsupported mode; ' +
                           'attempting to create depth+stencil buffer instead')
            self._stencilbuffer_attached = self._depthbuffer_attached = True
            cgl.glBindFramebuffer(GL_FRAMEBUFFER, old_fid)
            self.create_fbo()
            return

        if status != GL_FRAMEBUFFER_COMPLETE:
            self.raise_exception('FBO Initialization failed', status)

        # clear the fbo
        if do_clear:
            self.clear_buffer()

        # unbind the framebuffer
        cgl.glBindFramebuffer(GL_FRAMEBUFFER, old_fid)

        cdef Matrix projection_mat = Matrix()
        projection_mat.view_clip(0.0, self._width, 0.0, self._height, -1.0, 1.0, 0)
        self.set_state('projection_mat', projection_mat)

    cpdef bind(self):
        '''Bind the FBO to the current opengl context.
        `Bind` mean that you enable the Framebuffer, and all the drawing
        operations will act inside the Framebuffer, until :meth:`release` is
        called.

        The bind/release operations are automatically called when you add
        graphics objects into it. If you want to manipulate a Framebuffer
        yourself, you can use it like this::

            self.fbo = FBO()
            self.fbo.bind()
            # do any drawing command
            self.fbo.release()

            # then, your fbo texture is available at
            print(self.fbo.texture)
        '''
        if self._is_bound:
            self.raise_exception('FBO already binded.')
        else:
            self._is_bound = 1

        # stack our fbo to the last binded fbo
        cdef GLint old_fid = 0
        if len(fbo_stack) == 0:
            # the very first time we're going to create it, fill with the
            # initial framebuffer
            cgl.glGetIntegerv(GL_FRAMEBUFFER_BINDING, &old_fid)
            fbo_stack.append(old_fid)
        fbo_stack.append(self.buffer_id)
        cgl.glBindFramebuffer(GL_FRAMEBUFFER, self.buffer_id)

        # if asked, push the viewport
        if self._push_viewport:
            cgl.glGetIntegerv(GL_VIEWPORT, <GLint *>self._viewport)
            cgl.glViewport(0, 0, self._width, self._height)

        # save stencil stack
        self._stencil_state = get_stencil_state()
        reset_stencil_state()

    cpdef release(self):
        '''Release the Framebuffer (unbind).
        '''
        if self._is_bound == 0:
            self.raise_exception('FBO cannot be released (not binded).')
        else:
            self._is_bound = 0

        # bind the latest fbo, or unbind it.
        fbo_stack.pop()
        cgl.glBindFramebuffer(GL_FRAMEBUFFER, fbo_stack[-1])

        # if asked, restore the viewport
        if self._push_viewport:
            cgl.glViewport(self._viewport[0], self._viewport[1],
                           self._viewport[2], self._viewport[3])

        # restore stencil stack
        restore_stencil_state(self._stencil_state)

    cpdef clear_buffer(self):
        '''Clear the framebuffer with the :attr:`clear_color`.

        You need to bind the framebuffer yourself before calling this
        method::

            fbo.bind()
            fbo.clear_buffer()
            fbo.release()

        '''
        cgl.glClearColor(self._clear_color[0], self._clear_color[1],
                         self._clear_color[2], self._clear_color[3])
        if self._depthbuffer_attached and self._stencilbuffer_attached:
            cgl.glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT |
                        GL_STENCIL_BUFFER_BIT)
        elif self._depthbuffer_attached:
            cgl.glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        elif self._stencilbuffer_attached:
            cgl.glClear(GL_COLOR_BUFFER_BIT | GL_STENCIL_BUFFER_BIT)
        else:
            cgl.glClear(GL_COLOR_BUFFER_BIT)

    cdef int apply(self) except -1:
        if self.flags & GI_NEEDS_UPDATE:
            self.bind()
            RenderContext.apply(self)
            self.release()
            self.flag_update_done()
        return 0

    cdef void reload(self) except *:
        # recreate the framebuffer, without deleting it. the deletion is not
        # handled by us.
        self.create_fbo()
        self.flag_update()
        # notify observers
        for callback in self.observers[:]:
            method = callback()
            if method is None:
                self.observers.remove(callback)
                continue
            method(self)

    def add_reload_observer(self, callback):
        '''Add a callback to be called after the whole graphics context has
        been reloaded. This is where you can reupload your custom data in GPU.

        .. versionadded:: 1.2.0

        :Parameters:
            `callback`: func(context) -> return None
                The first parameter will be the context itself
        '''
        self.observers.append(WeakMethod(callback))

    def remove_reload_observer(self, callback):
        '''Remove a callback from the observer list, previously added by
        :meth:`add_reload_observer`.

        .. versionadded:: 1.2.0

        '''
        for cb in self.observers[:]:
            method = cb()
            if method is None or method is callback:
                self.observers.remove(cb)
                continue


    @property
    def size(self):
        '''Size of the framebuffer, in (width, height) format.

        If you change the size, the framebuffer content will be lost.
        '''
        return (self._width, self._height)

    @size.setter
    def size(self, x):
        cdef int w, h
        w, h = x
        if w == self._width and h == self._height:
            return
        self._width, self._height = x
        self.delete_fbo()
        self.create_fbo()
        self.flag_data_update()

    @property
    def clear_color(self):
        '''Clear color in (red, green, blue, alpha) format.
        '''
        return (self._clear_color[0],
                self._clear_color[1],
                self._clear_color[2],
                self._clear_color[3])

    @clear_color.setter
    def clear_color(self, x):
        x = list(x)
        if len(x) != 4:
            raise Exception('clear_color must be a list/tuple of 4 entry.')
        self._clear_color[0] = x[0]
        self._clear_color[1] = x[1]
        self._clear_color[2] = x[2]
        self._clear_color[3] = x[3]

    @property
    def texture(self):
        '''Return the framebuffer texture
        '''
        return self._texture

    @property
    def pixels(self):
        '''Get the pixels texture, in RGBA format only, unsigned byte. The
        origin of the image is at bottom left.

        .. versionadded:: 1.7.0
        '''
        w, h = self._width, self._height
        self.bind()
        data = py_glReadPixels(0, 0, w, h, GL_RGBA, GL_UNSIGNED_BYTE)
        self.release()
        return data

    cpdef get_pixel_color(self, int wx, int wy):
        """Get the color of the pixel with specified window
        coordinates wx, wy. It returns result in RGBA format.

        .. versionadded:: 1.8.0
        """
        if wx > self._width or wy > self._height:
            # window coordinates should not exceed the
            # frame buffer size
            return (0, 0, 0, 0)
        self.bind()
        data = py_glReadPixels(wx, wy, 1, 1, GL_RGBA, GL_UNSIGNED_BYTE)
        self.release()

        return [ord(i) if PY2 else i for i in data]



'''
Context management
==================

.. versionadded:: 1.2.0

This class manages a registry of all created graphics instructions. It has
the ability to flush and delete them.

You can read more about Kivy graphics contexts in the :doc:`api-kivy.graphics`
module documentation. These are based on
`OpenGL graphics contexts <http://www.opengl.org/wiki/OpenGL_Context>`_.
'''

__all__ = ('Context',)

include "../include/config.pxi"

from cpython.array cimport array
import gc
from os import environ
from weakref import ref
from kivy.graphics.instructions cimport Canvas
from kivy.graphics.texture cimport Texture, TextureRegion
from kivy.graphics.vbo cimport VBO, VertexBatch
from kivy.logger import Logger
from kivy.clock import Clock

from kivy.graphics.cgl cimport *

from kivy.weakmethod import WeakMethod
from time import time
from kivy.cache import Cache

cdef Context context = None

cdef class Context:
    """
    The Context class manages groups of graphics instructions. It can also be used to manage
    observer callbacks. See :meth:`add_reload_observer` and :meth:`remove_reload_observer`
    for more information.
    """
    def __init__(self):
        self.observers = []
        self.observers_before = []
        self.l_texture = []
        self.l_canvas = []
        self.l_fbo = []
        self.flush()

    def trigger_gl_dealloc(self):
        # see https://github.com/kivy/kivy/issues/5986
        # Somehow, the try/except fixes that issue. Calling `self.gl_dealloc()`
        # directly also caused the `Error in sys.excepthook:`, so it's not just
        # the Clock. Perhaps this is called after everything has been deleted
        # and cython objects destroyed, which like a bug in cython/python
        try:
            Clock.schedule_del_safe(self.gl_dealloc)
        except Exception:
            pass

    cpdef void flush(self):
        gc.collect()
        self.lr_texture = array('I')
        self.lr_canvas = []
        self.lr_vbo = array('I')
        self.lr_fbo_rb = array('I')
        self.lr_fbo_fb = array('I')
        self.lr_shadersource = array('I')
        self.lr_shader = []

    cdef void register_texture(self, Texture texture):
        self.l_texture.append(ref(texture, self.l_texture.remove))

    cdef void register_canvas(self, Canvas canvas):
        self.l_canvas.append(ref(canvas, self.l_canvas.remove))

    cdef void register_fbo(self, Fbo fbo):
        self.l_fbo.append(ref(fbo, self.l_fbo.remove))

    cdef void dealloc_texture(self, Texture texture):
        cdef array arr
        if texture._nofree or texture.__class__ is TextureRegion:
            return
        if texture.id > 0:
            arr = self.lr_texture
            arr.append(texture.id)
            self.trigger_gl_dealloc()

    cdef void dealloc_vbo(self, VBO vbo):
        cdef array arr
        if vbo.have_id():
            arr = self.lr_vbo
            arr.append(vbo.id)
            self.trigger_gl_dealloc()

    cdef void dealloc_vertexbatch(self, VertexBatch batch):
        cdef array arr
        if batch.have_id():
            arr = self.lr_vbo
            arr.append(batch.id)
            self.trigger_gl_dealloc()

    cdef void dealloc_shader(self, Shader shader):
        if shader.program == 0:
            return
        cdef int vs_id = -1
        cdef int fs_id = -1
        self.lr_shader.append((shader.program, vs_id, fs_id))

    cdef void dealloc_shader_source(self, int shader):
        cdef array arr
        if shader == -1:
            return
        arr = self.lr_shadersource
        arr.append(shader)
        self.trigger_gl_dealloc()

    cdef void dealloc_fbo(self, Fbo fbo):
        cdef array arr_fb
        cdef array arr_rb
        if fbo.buffer_id != 0:
            arr_fb = self.lr_fbo_fb
            arr_fb.append(fbo.buffer_id)
            self.trigger_gl_dealloc()
        if fbo.depthbuffer_id != 0 or fbo.stencilbuffer_id != 0:
            arr_rb = self.lr_fbo_rb
            arr_rb.append(fbo.depthbuffer_id or fbo.stencilbuffer_id)
            # no need to trigger, depthbuffer required absolutely a buffer.

    def add_reload_observer(self, callback, before=False):
        '''(internal) Add a callback to be called after the whole graphics context has
        been reloaded. This is where you can reupload your custom data into the
        GPU.

        :Parameters:
            `callback`: func(context) -> return None
                The first parameter will be the context itself
            `before`: boolean, defaults to False
                If True, the callback will be executed before all the
                reloading processes. Use it if you want to clear your cache for
                example.

        .. versionchanged:: 1.4.0
            `before` parameter added.
        '''
        if before:
            self.observers_before.append(WeakMethod(callback))
        else:
            self.observers.append(WeakMethod(callback))

    def remove_reload_observer(self, callback, before=False):
        '''(internal) Remove a callback from the observer list previously added by
        :meth:`add_reload_observer`.
        '''
        lst = self.observers_before if before else self.observers
        for cb in lst[:]:
            method = cb()
            if method is None or method is callback:
                lst.remove(cb)
                continue

    def reload(self):
        cdef VBO vbo
        cdef VertexBatch batch
        cdef Texture texture
        cdef Shader shader
        cdef Canvas canvas

        # call reload observers that want to do something after a whole gpu
        # reloading.
        for callback in self.observers_before[:]:
            method = callback()
            if method is None:
                self.observers_before.remove(callback)
                continue
            method(self)

        # mark all the texture to not delete from the previous reload as to
        # delete now.
        for item in self.l_texture[:]:
            texture = item()
            if texture is None:
                continue
            if texture._nofree == 1:
                texture._nofree = 0
                self.l_texture.remove(item)

        image_objects = Cache._objects['kv.image']
        Cache.remove('kv.loader')
        Cache.remove('kv.image')
        Cache.remove('kv.shader')

        # For texture cache, save the objects. We need to clean the cache as the
        # others to prevent of using it during the reloading part.
        # We'll restore the object later.
        texture_objects = Cache._objects['kv.texture']
        Cache.remove('kv.texture')

        start = time()
        Logger.info('Context: Reloading graphics data...')
        Logger.debug('Context: Collect and flush all garbage')
        self.flush()

        # First step, prevent double loading by setting everything to -1
        # We do this because texture might be loaded in separate texture at first,
        # then merged from the cache cause of the same source
        Logger.debug('Context: Reload textures')
        cdef list l = self.l_texture[:]
        for item in l:
            texture = item()
            if texture is None:
                continue
            Logger.trace('Context: unset texture id %r' % texture)
            texture._id = -1

        # First time, only reload base texture
        for item in l:
            texture = item()
            if texture is None or isinstance(texture, TextureRegion):
                continue
            Logger.trace('Context: >> reload base texture %r' % texture)
            texture.reload()
            Logger.trace('Context: << reload base texture %r' % texture)

        # Second time, update texture region id
        for item in l:
            texture = item()
            if texture is None or not isinstance(texture, TextureRegion):
                continue
            Logger.trace('Context: >> reload region texture %r' % texture)
            texture.reload()
            Logger.trace('Context: << reload region texture %r' % texture)

        # Restore texture cache
        texture_objects.update(Cache._objects['kv.texture'])
        Cache._objects['kv.texture'] = texture_objects
        image_objects.update(Cache._objects['kv.image'])
        Cache._objects['kv.image'] = image_objects

        gc_objects = gc.get_objects()[:]
        Logger.debug('Context: Reload vbos')
        for item in gc_objects:
            if isinstance(item, VBO):
                vbo = item
                Logger.trace('Context: reloaded %r' % vbo)
                vbo.reload()
        Logger.debug('Context: Reload vertex batchs')
        for item in gc_objects:
            if isinstance(item, VertexBatch):
                batch = item
                Logger.trace('Context: reloaded %r' % batch)
                batch.reload()
        Logger.debug('Context: Reload shaders')
        for item in gc_objects:
            if isinstance(item, Shader):
                shader = item
                Logger.trace('Context: reloaded %r' % shader)
                shader.reload()
        Logger.debug('Context: Reload canvas')
        for item in self.l_canvas[:]:
            canvas = item()
            if canvas is not None:
                Logger.trace('Context: reloaded %r' % item())
                canvas.reload()

        # call reload observers that want to do something after a whole gpu
        # reloading.
        for callback in self.observers[:]:
            method = callback()
            if method is None:
                self.observers.remove(callback)
                continue
            method(self)

        cgl.glFinish()
        dt = time() - start
        Logger.info('Context: Reloading done in %2.4fs' % dt)

    def flag_update_canvas(self):
        cdef Canvas canvas
        for item in self.l_canvas:
            canvas = item()
            if canvas:
                canvas.flag_update()

    def gl_dealloc(self, *largs):
        # dealloc all gl resources asynchronously
        cdef GLuint i, j
        cdef array arr

        if len(self.lr_vbo):
            Logger.trace('Context: releasing %d vbos' % len(self.lr_vbo))
            arr = self.lr_vbo
            cgl.glDeleteBuffers(<GLsizei>len(self.lr_vbo), arr.data.as_uints)
            del self.lr_vbo[:]
        if len(self.lr_texture):
            Logger.trace('Context: releasing %d textures: %r' % (
                len(self.lr_texture), self.lr_texture))
            arr = self.lr_texture
            cgl.glDeleteTextures(<GLsizei>len(self.lr_texture), arr.data.as_uints)
            del self.lr_texture[:]
        if len(self.lr_fbo_fb):
            Logger.trace('Context: releasing %d framebuffer fbos' % len(self.lr_fbo_fb))
            arr = self.lr_fbo_fb
            cgl.glDeleteFramebuffers(<GLsizei>len(self.lr_fbo_fb), arr.data.as_uints)
            del self.lr_fbo_fb[:]
        if len(self.lr_fbo_rb):
            Logger.trace('Context: releasing %d renderbuffer fbos' % len(self.lr_fbo_fb))
            arr = self.lr_fbo_rb
            cgl.glDeleteRenderbuffers(<GLsizei>len(self.lr_fbo_rb), arr.data.as_uints)
            del self.lr_fbo_rb[:]
        if len(self.lr_shadersource):
            Logger.trace('Context: releasing %d shader sources' % len(self.lr_shadersource))
            arr = self.lr_shadersource
            for i in self.lr_shadersource:
                cgl.glDeleteShader(i)
            del self.lr_shadersource[:]
        if len(self.lr_shader):
            Logger.trace('Context: releasing %d shaders' % len(self.lr_shader))
            for program, vs_id, fs_id in self.lr_shader:
                if vs_id != -1:
                    cgl.glDetachShader(program, vs_id)
                if fs_id != -1:
                    cgl.glDetachShader(program, fs_id)
                cgl.glDeleteProgram(program)
            del self.lr_shader[:]


cpdef Context get_context():
    global context
    if context is None:
        context = Context()
    return context
